<?php

set_time_limit(0);
//ini_set('error_reporting', E_ALL & ~E_NOTICE);
@ini_set('error_reporting', E_ALL & ~E_NOTICE & ~E_DEPRECATED);

// Define app name
define('APP_NAME','ISPCopy');
define('APP_VERSION', '1.1.0');

/* path definitions */
define('BASE_PATH', realpath(dirname(__FILE__)));
define('INC_PATH', BASE_PATH . '/lib');


// Set the log level
$log_level = 0; // 0 = debug, 1 = warning, 2 = error

// Include the ISPConfig 3 config file and the app inc file
require '/usr/local/ispconfig/server/lib/config.inc.php';
require '/usr/local/ispconfig/server/lib/app.inc.php';

// Load some modules
$app->uses('ini_parser,file,services,getconf,system');

// Include the libraries
require INC_PATH.'/lib.php';
require INC_PATH.'/isp_copy.lib.php';
require INC_PATH.'/ssh.lib.php';

$icp = new isp_copy();
// Check if we are on a ISPConfig 3 system
if(!is_file('/usr/local/ispconfig/server/lib/config.inc.php')) $icp->end_copy('No ISPConfig 3 found on this server.');

//$icp->migrate_mysql_db('dbispconfig');
//$icp->end_copy();

// Check prerequisites
if(!$app->system->is_installed('rsync')) $icp->end_copy('Please install rsync and try again.');

// Load the server config and check if it is accessible
$app->uses('getconf,ini_parser');
$server_config = $app->getconf->get_server_config($conf['server_id']);

if(!is_array($server_config) && count($server_config) == 0) $icp->end_copy('Unable to read ISPConfig 3 server configuration.');

//** Get commandline options
$cmd_opt = getopt('', array('autoconf::'));

//** Load autorun file
if(isset($cmd_opt['autoconf']) && is_file($cmd_opt['autoconf'])) {
	if(is_file('autoconf.ini')) {
		$tmp = $icp->ini_to_array(file_get_contents('autoconf.ini'));
	} else {
		$tmp = $icp->ini_to_array(file_get_contents($cmd_opt['autoconf']));
	}
	$autoconf = $tmp['settings'];
	define('AUTORUN', true);
} else {
	$autorun = array();
	define('AUTORUN', false);
}

echo "
  _____  _____ _____   _____                  
 |_   _|/ ____|  __ \ / ____|                 
   | | | (___ | |__) | |     ___  _ __  _   _ 
   | |  \___ \|  ___/| |    / _ \| '_ \| | | |
  _| |_ ____) | |    | |___| (_) | |_) | |_| |
 |_____|_____/|_|     \_____\___/| .__/ \__, |
                                 | |     __/ |
                                 |_|    |___/ 
";

$icp->swriteln("Check for newest version...");
$package_updated = false;
do {
    exec("apt-get update > /dev/null 2>&1", $update_output, $update_result);
    $package_updated = ($update_result == 0);
} while(!$package_updated);

exec("apt-cache policy th-ispcopy | grep \"Installed\" | cut -d':' -f 2", $app_version);
exec("apt-cache policy th-ispcopy | grep \"Candidate\" | cut -d':' -f 2", $app_candidate);
$icp->swriteln('');
$icp->swriteln('Welcome to the ISPConfig Copy Tool '.trim($app_version[0]).'.');
if(trim($app_version[0]) != trim($app_candidate[0])) {
    $die = $icp->simple_query("Version " . trim($app_candidate[0]) . " is available but not installed. Do you like to cancel and retry with updated version?",
        ["y", "n"],
        "y");
    if($die == "y") {
        die();
    }
}
$icp->swriteln('');
$icp->swriteln('* This tool will help you to move an ISPConfig installation to a new server.');
$icp->swriteln('* The target and source server must run the same Linux Distribution.');
$icp->swriteln('* The source server does not get modified, but it must have enough free space');
$icp->swriteln('  in its temporary folder to store the dump of one MySQL database at a time.');
$icp->swriteln('* Ensure that you make a backup of your source and target systems before you start.');
$icp->swriteln('');
$icp->swriteln('');
$icp->swriteln($icp->text_server_info);
$icp->swriteln('');
$icp->swriteln('');


$icp->swriteln('Please confirm that you run the ISPCopy on the source (old) server and not on the target.');
$result = $icp->simple_query('Do you like to start now?', array('y','n'), 'y', 'start');
if($result == 'n') {
	$icp->swriteln('We stop here.');
	$icp->end_copy();
}


// Rotate the logfile and set the loglevel
$icp->setLogLevel($log_level);
if(is_file($icp->getLogFile())) rename($icp->getLogFile(),$icp->getLogFile().'~');

// set secure permissions for log file
touch($icp->getLogFile());
chmod($icp->getLogFile(), 0600);

// Get the target server details
$target_server_hostname = $icp->free_query('Target Server hostname', '', 'target_server_hostname');
$target_server_ip = $icp->free_query('Target Server IP', '', 'target_server_ip');
$target_server_port = intval($icp->free_query('Target Server SSH port', '22', 'target_server_port'));
$timme_user = $icp->free_query("Timme User", $icp->getTimmeUser(), "timme_user", true);

// Regex-Prüfungen für Hostname und IP
//if(!filter_var($target_server_hostname, FILTER_VALIDATE_DOMAIN, FILTER_FLAG_HOSTNAME)){ // erst unterstützt ab PHP 7.0
if(!preg_match('/^[\w\.\-]{2,255}\.[a-zA-Z\-]{2,30}$/', $target_server_hostname)){
	$icp->swriteln($icp->lng('No valid target server hostname specified.'));
	$icp->end_copy();
}
if(!filter_var($target_server_ip, FILTER_VALIDATE_IP) || empty($target_server_ip)){
	$icp->swriteln($icp->lng('No valid target server IP specified.'));
	$icp->end_copy();
}

$icp->target_server_hostname = $target_server_hostname;
$dns_ip = gethostbyname($icp->target_server_hostname);
if($dns_ip == $target_server_ip){
	$icp->target_server_ip = $target_server_ip;
} else {
	$icp->swriteln($icp->lng('Hostname '.$target_server_hostname.' resolves to another IP ('.$dns_ip.') than the one specified ('.$target_server_ip.').'));
	$icp->end_copy();
}

//$icp->copy_ssh_key($timme_user);


// Create SSH connection
$ssh = new ssh_lib($target_server_hostname,$target_server_port);
if(!$ssh->testConnect()) {
    $ssh->installSSHCert($timme_user);
}
$ssh->setLogLevel($log_level);

// Test SSH connection as root user without password
//if(!$ssh->testConnect()) {
//	$icp->swriteln($icp->lng('Creating SSH key-based login for the root user.'));
//	$ssh->installSSHCert();
//}

// Quit the program if no SSH login is possible.
if(!$ssh->testConnect()) {
	$icp->swriteln($icp->lng('Unable to create keyless SSH login.'));
	$icp->end_copy();
}
$ssh->create_deafaults_extra_file();

// DBD Test (you know what this means)
$testpath = '/tmp/'.uniqid();
touch($testpath);
if($ssh->isFile($testpath) === true) {
	unlink($testpath);
	$icp->swriteln($icp->lng('SSH Login points to local server.'));
	$icp->end_copy();
} else {
	unlink($testpath);
}

if(!isset($argv) || !is_array($argv) || !in_array('--data-only', $argv, true)) {
    $mode = $icp->checkRunMode();
    if($mode == "second") {
        if(!isset($argv) || !is_array($argv)) {
            $argv = [];
        }
        if(!in_array('--data-only', $argv, true)) {
            array_push($argv, "--data-only");
        }
    }
    if($mode == "") {
        $icp->end_copy("No valid run mode for isp copy.");
    }
}

if(!empty($icp->master_server)){
    // Create SSH connection to master server
    $master_ssh = new ssh_lib($icp->master_server, 22);
    if(!$master_ssh->testConnect()) {
        $master_ssh->installSSHCert($timme_user);
    }
    $master_ssh->setLogLevel($log_level);

	// Quit the program if no SSH login is possible.
	if(!$master_ssh->testConnect()) {
		$icp->swriteln($icp->lng('Unable to create keyless SSH login on the master server.'));
		$icp->end_copy();
	}

    $master_ssh->create_deafaults_extra_file();

	// DBD Test (you know what this means)
	$testpath = '/tmp/'.uniqid();
	touch($testpath);
	if($master_ssh->isFile($testpath) === true) {
		unlink($testpath);
		$icp->swriteln($icp->lng('SSH Login points to local server instead of master server.'));
		$icp->end_copy();
	} else {
		unlink($testpath);
	}
}

if(isset($argv) && is_array($argv) && in_array('--data-only', $argv, true)) {
    $icp->askQuestions("second");

    activateMaintenance($icp);

	// Migrate the MySQL databases
	$icp->swriteln($icp->lng('Copying databases...'));
	$icp->migrate_databases();

	// Migrate files
	$icp->swriteln($icp->lng('Copying Files...'));
	$icp->migrate_files();

    if(empty($icp->master_server)) {
// Server-Wartungsmodus auf Quellserver deaktivieren
        $icp->set_server_maintenance(false, 'source');
    }

    if($mode == "second") {
        $icp->replaceConfig("master");
        $icp->replaceConfig();
// Adjust config.inc.php on target server
        $icp->swriteln($icp->lng('Adjusting config.inc.php on target server...'));
        $icp->adjust_target_config();

// Replace IP Addresses
        $icp->replace_ip();
        sleep(10);

        if(!empty($icp->master_server)){
// if this is a slave server, change MySQL access permissions on master server
            $icp->swriteln($icp->lng('Changing MySQL access permissions on master server...'));
            $icp->master_change_mysql_access();

// if this is a slave server, disable ISPConfig interface on target server
            $icp->swriteln($icp->lng('Disabling ISPConfig interface on target server...'));
            $icp->target_disable_ispconfig_interface();
        }
// if this is a master server, transfer MySQL slave access permissions to target server and adjust master server on slaves' config.inc.php files
        $icp->new_master_adjust_slaves();
// do a resync on the remote server
        if(!empty($icp->master_server)) {
            $icp->resetSlavesSysDatalogOnSecondTransfer($mode);
        }
        $icp->target_resync();
        $icp->copySSL();
    }
    doPostWork();

    $icp->end_copy("Copy finished successfull.");
}

// Check if target server is empty
$empty = $icp->target_is_empty();
if(!$empty){
	$icp->swriteln($icp->lng('Target server is NOT empty.'));
	$icp->end_copy();
}

$icp->askQuestions($mode);

activateMaintenance($icp);

// Get default vhost from target server
$icp->swriteln($icp->lng('Getting default vhost from target server...'));
$icp->get_target_default_vhost();

// Get PHP versions from target server
$icp->swriteln($icp->lng('Getting PHP versions from target server...'));
$icp->get_target_php_versions();
//$icp->migrate_mysql_db('dbispconfig');
//$icp->end_copy();

// Delete Vhost with ID 1 on target server
$icp->swriteln($icp->lng('Deleting default vhost on target server...'));
$icp->delete_target_default_vhost();

// Server-Wartungsmodus auf Zielserver aktivieren
$icp->set_server_maintenance(true, 'target');

// Migrate the SSH users and groups
$icp->swriteln($icp->lng('Copying system users...'));
$icp->migrate_ssh_users();

$icp->swriteln($icp->lng('Copying databases...'));
$icp->migrate_databases();

$slave_as_first = false;
if($icp->server_role != "single" && $mode == "first") {
    $icp->master_server = null;
    $icp->server_role = "single";
    $slave_as_first = true;
}

// Adjust config.inc.php on target server
$icp->swriteln($icp->lng('Adjusting config.inc.php on target server...'));
$icp->adjust_target_config();

// Migrate files
$icp->swriteln($icp->lng('Copying Files...'));
$icp->migrate_files();

// Replace IP Addresses
$icp->replace_ip();
sleep(10);

// Restart services on remote system
//$icp->swriteln($icp->lng('Restarting services on target system...'));
//$icp->restart_services();

// if this is a slave server, change MySQL access permissions on master server
if(!empty($icp->master_server)){
	$icp->swriteln($icp->lng('Changing MySQL access permissions on master server...'));
	$icp->master_change_mysql_access();
}
// if this is a slave server, disable ISPConfig interface on target server
if(!empty($icp->master_server)){
	$icp->swriteln($icp->lng('Disabling ISPConfig interface on target server...'));
	$icp->target_disable_ispconfig_interface();
}

if(empty($icp->master_server)) {
// Server-Wartungsmodus auf Quellserver deaktivieren
    $icp->set_server_maintenance(false, 'source');
}

//// if this is a master server, transfer MySQL slave access permissions to target server and adjust master server on slaves' config.inc.php files
//$icp->new_master_adjust_slaves();

// start docker containers on remote server
$icp->swriteln($icp->lng('Starting docker containers on target system...'));
$icp->start_docker_containers();

// copy Docker databases to remote server
$icp->swriteln($icp->lng('Migrating Docker databases to target system...'));
$icp->migrate_docker_databases();

// do a resync on the remote server
if($slave_as_first) {
    $icp->resetSlavesSysDatalogOnSecondTransfer($mode);
}
$icp->target_resync();

doPostWork();

$icp->end_copy("Copy finished successfull. Press Enter to close.");


function activateMaintenance($icp) {
    $icp->stopSysCrons();
    $icp->stopUserCrons();
    $icp->disableServerShOnTarget();
// Interface-Wartungsmodus aktivieren
    $icp->set_interface_maintenance(true);
// Wartungsmodus für alle Vhosts des Servers aktivieren
    $icp->activate_user_maintenance();
// Server-Wartungsmodus auf Quellserver aktivieren
    $icp->set_server_maintenance(true, 'source');
}

function doPostWork() {
    global $icp, $ssh;

// Interface-Wartungsmodus deaktivieren
    $icp->set_interface_maintenance(false);
// Server-Wartungsmodus auf Zielserver deaktivieren
    $icp->set_server_maintenance(false, 'target');

    $icp->runServerSh("target");

// automation of manual tasks
    $icp->copy_crons_to_tmp();
//    $target_updated = false;
    $target_updated = $icp->activateSNI();
    if($target_updated) {
        $target_updated = $icp->createServerName();
    }
    $icp->setProxy();
    $icp->setMailRouting();

    $replacement_result = $icp->replaceServernameInAliasdomains();
    if(is_bool($replacement_result)) {
        $target_updated = $replacement_result || $target_updated;
    }

    $icp->setDockerRootPw();

// Wartungsmodus für vhosts des Server auf dem Zielsystem deaktivieren
    $target_updated = $icp->disable_user_maintenance() || $target_updated;
    $target_updated = $icp->toggleUserCronsOnTarget() || $target_updated;

    if($target_updated) {
        $icp->runServerSh("target");
    }

    if(is_string($replacement_result)) {
        if($icp->simple_query("Shall the following replacement be executed on the target:\n" . $replacement_result,
            ["y", "n"],
            "y") == "y") {
            $ssh->exec($replacement_result);
        }
    }

    $icp->replaceServerId();

    $icp->enableServerShOnTarget();

    if($icp->isOnTelekomBlacklist()) {
        $icp->swriteln("");
        $icp->swriteln("Telekom blacklisted the server {$icp->target_server_ip} ({$icp->target_server_hostname}).");
    }

    $icp->printSuccessMessage();
}
