<?php

class isp_copy extends lib {
	
	private $local_mysql_root_user = 'root';
	private $local_mysql_root_password = '';
	private $local_mysql_host = 'localhost';
	private $remote_mysql_root_user = 'root';
	private $remote_mysql_root_password = '';
	private $remote_mysql_host = 'localhost';
	private $web1 = array();
	private $target_php_versions = array();
	
	var $server_role = 'single'; // single|master|slave
	private $server_id;
	private $mirror_server_id = 0; // falls der Slave ein Mirror ist, steht hier die server_id des Masters; falls der Slave kein Mirror ist, bleibt die mirror_server_id 0.
	var $master_server;
	private $master_mysql_root_user = 'root';
	private $master_mysql_root_password = '';
	private $master_mysql_host = 'localhost';
	
	var $local_server_hostname;
	var $local_server_ip;
	var $target_server_hostname;
	var $target_server_ip;
	
	private $maintenance_file = '/usr/local/ispconfig/interface/.maintenance';
	private $server_maintenance_file = '/usr/local/ispconfig/server/.maintenance';

    var $text_server_info = "";
	
	public function __construct() {
		global $app, $conf;
		
		$this->server_id = $conf['server_id'];
		$server = $app->db->queryOneRecord("SELECT * FROM server WHERE server_id = ?", $this->server_id);
		$this->local_server_hostname = $server['server_name'];
		$this->local_server_ip = gethostbyname($this->local_server_hostname);

		if ($app->is_slave()) {
			$this->server_role = 'slave';
			$this->master_server = $conf['dbmaster_host'];
			if(empty($this->master_server)) $this->error('This server is a slave, but no master server was found.');
            $this->text_server_info = 'The current server is a slave server. The master server is '.$this->master_server.'.';
			
			if($server['mirror_server_id'] > 0){
				$this->mirror_server_id = $server['mirror_server_id'];
                $this->text_server_info = 'The current server is a mirror server. The master server has the server id '.$this->mirror_server_id.'.';
			}
		} else {
			$other_servers = $app->db->queryAllRecords("SELECT * FROM server WHERE server_id != ?", $this->server_id);
			if(!empty($other_servers) && is_array($other_servers)){
				$this->server_role = 'master';
                $this->text_server_info = 'The current server is a master server.';
			} else {
                $this->text_server_info = 'The current server is a single server.';
			}
		}

        if(!is_dir($this->getLogDir())) {
            mkdir($this->getLogDir());
        }

        if(!file_exists($this->mysql_defaults_extra_file)) {
            if(file_exists('/usr/local/ispconfig/server/lib/mysql_clientdb.conf')) {
                include '/usr/local/ispconfig/server/lib/mysql_clientdb.conf';

                $this->local_mysql_root_user = $clientdb_user;
                $this->local_mysql_root_password = $clientdb_password;
                $this->local_mysql_host = $clientdb_host;

                $content = <<<TXT
[client]
host=$clientdb_host
user=$clientdb_user
password=$clientdb_password
TXT;
                file_put_contents($this->mysql_defaults_extra_file, $content);
            }
        }
	}
	
	public function sread() {
		$input = fgets(STDIN);
		return rtrim($input);
	}

	public function swrite($text = '') {
		echo $text;
	}

	public function swriteln($text = '') {
		echo $text."\n";
	}
	
	public function lng($text) {
		return $text;
	}

	public function error($msg) {
		// Wartungsmodus deaktivieren
		$this->set_maintenance(false);
		die('ERROR: '.$msg."\n");
	}

	public function warning($msg) {
		echo 'WARNING: '.$msg."\n";
	}
	
	public function log($message, $priority = 0) {
		if($priority >= $this->getLogLevel()) {
			// Replace passwords in log
			$message = str_replace($this->local_mysql_root_password,'xxxxxxxxx',$message);
			$message = str_replace($this->remote_mysql_root_password,'xxxxxxxxx',$message);
			// Write log line
			file_put_contents($this->getLogFile(),date('Y-m-d H:i:s').' - '.$message."\n",FILE_APPEND);
		}
	}

    private function writeInputLog($key, $value, $section = "settings") {
        $log = [$section => []];
        if(is_file($this->getInputLogFile())) {
            $log = $this->ini_to_array(file_get_contents($this->getInputLogFile()));
        }
        if(!array_key_exists($section, $log)) {
            $log[$section] = [];
        }
        $log[$section][$key] = $value;
        file_put_contents($this->getInputLogFile(), $this->array_to_ini($log));
    }

    private function getInputLogEntry($key, $section = "settings") {
        $value = "";
        if(is_file($this->getInputLogFile())) {
            $log = $this->ini_to_array(file_get_contents($this->getInputLogFile()));
            if(array_key_exists($section, $log) && array_key_exists($key, $log[$section])) {
                $value = $log[$section][$key];
            }
        }
        return $value;
    }

    private function getInputLogSection($section = "settings") {
        $result = [];
        if(is_file($this->getInputLogFile())) {
            $log = $this->ini_to_array(file_get_contents($this->getInputLogFile()));
            if(array_key_exists($section, $log)) {
                $result = $log[$section];
            }
        }
        return $result;
    }

    public function checkRunMode() {
        $mode = $this->getInputLogEntry("mode");
        if(empty($mode)) {
            return $this->simple_query("How to start ISPCopy (single = single-/full transfer, first = first part of double transfer)?", ["single", "first"], "single", "mode");
        } else {
            if($this->simple_query("Start second part of double transfer?", ["y", "n"], "y") == "y") {
                //$this->writeInputLog("mode", "second");
                return "second";
            } else {
                $this->writeInputLog("mode", "");
                return $this->checkRunMode();
            }
        }
    }
	
	public function simple_query($query, $answers, $default, $name = '', $ignore_input_file = false) {
		global $autoconf;
		$finished = false;
		do {
            if(!$ignore_input_file && !empty($name) && !empty($this->getInputLogEntry($name))) {
                $default = $this->getInputLogEntry($name);
            }
			if($name != '' && !empty($autoconf) && !empty($autoconf[$name])) {
				if($autoconf[$name] == 'default') {
					$input = $default;
				} else {
					$input = $autoconf[$name];
				}
			} else {
				$answers_str = implode(', ', $answers);
				$this->swrite($this->lng($query).' ('.$answers_str.') ['.$default.']: ');
				$input = $this->sread();
			}

			//* Stop the program
			if($input == 'quit') {
				$this->swriteln($this->lng("Program terminated by user.\n"));
				$this->log('Program stopped with -quit-');
				die();
			}

			//* Select the default
			if($input == '') {
				$answer = $default;
				$finished = true;
			}

			//* Set answer id valid
			if(in_array($input, $answers)) {
				$answer = $input;
				$finished = true;
			}

		} while ($finished == false);
		$this->swriteln();
		$this->log('Query: '.$query.' -- Answer: '.$answer);
        if(!empty($name)) {
            $this->writeInputLog($name, $answer);
        }
		return $answer;
	}

	public function free_query($query, $default, $name = '', $ignore_input_file = false) {
		global $autoconf;
        if(!$ignore_input_file && !empty($name) && !empty($this->getInputLogEntry($name))) {
            $default = $this->getInputLogEntry($name);
        }
		if($name != '' && !empty($autoconf) && !empty($autoconf[$name])) {
			if($autoconf[$name] == 'default') {
				$input = $default;
			} else {
				$input = $autoconf[$name];
			}
		} else {
			$this->swrite($this->lng($query).' ['.$default.']: ');
			$input = $this->sread();
		}

		//* Stop the program
		if($input == 'quit') {
			$this->swriteln($this->lng("Program terminated by user.\n"));
			die();
		}
		$input = trim($input);
		$answer =  ($input == '') ? $default : $input;
		$this->swriteln();
		$this->log('Query: '.$query.' -- Answer: '.$answer);
        if(!empty($name)) {
            $this->writeInputLog($name, $answer);
        }
		return $answer;
	}
	
	//* Converts a ini string to array
	public function ini_to_array($ini) {
		$config = array();
		$ini = str_replace("\r\n", "\n", $ini);
		$lines = explode("\n", $ini);
		foreach($lines as $line) {
			$line = trim($line);
			if($line != '') {
				if(preg_match("/^\[([\w\d_]+)\]$/", $line, $matches)) {
					$section = strtolower($matches[1]);
				} elseif(preg_match("/^([\w\d_\.\:]+)=(.*)$/", $line, $matches) && $section != null) {
					$item = trim($matches[1]);
					$config[$section][$item] = trim($matches[2]);
				}
			}
		}
		return $config;
	}

	//* Converts a config array to a string
	public function array_to_ini($config_array = '') {
		if($config_array == '') $config_array = $this->config;
		$content = '';
		foreach($config_array as $section => $data) {
			$content .= "[$section]\n";
			foreach($data as $item => $value) {
				if($item != ''){
					$content .= "$item=$value\n";
				}
			}
			$content .= "\n";
		}
		return $content;
	}
	
	// Reads in a Linux passwd and shadow file and returns the information as an array
	private function passwd_file_to_array($passwd_file,$shadow_file = '') {
		$out = array();
		$pw_file_array = file($passwd_file,FILE_IGNORE_NEW_LINES);
		foreach($pw_file_array as $line) {
			$u = explode(':',$line);
			$username = $u[0];
			$out[$username] = array(
				'user' => $u[0],
				'pw' => $u[1],
				'uid' => $u[2],
				'gid' => $u[3],
				'name' => $u[4],
				'home' => $u[5],
				'shell' => $u[6]
			);
            $uid = $u[2];
            $out['_'.$uid] = $username;
		}
		if($shadow_file != '') {
			$shadow_file_array = file($shadow_file,FILE_IGNORE_NEW_LINES);
			foreach($shadow_file_array as $line) {
				$s = explode(':',$line);
				$username = $s[0];
				$out[$username]['password'] = $s[1];
			}
		}
		return $out;
	}
	
	// Reads in a Linux group file and returns the information as an array
	private function group_file_to_array($group_file) {
		$out = array();
		$group_file_array = file($group_file,FILE_IGNORE_NEW_LINES);
		foreach($group_file_array as $line) {
			$g = explode(':',$line);
			$groupname = $g[0];
			$out[$groupname] = array(
				'group' => $g[0],
				'pw' => $g[1],
				'gid' => $g[2],
				'members' => $g[3],
			);
		}
		$gid = $g[2];
		$out['_'.$gid] = $groupname;
		return $out;
	}
	
	// function to get available PHP versions on the remote server
	public function get_target_php_versions() {
		global $ssh;
		
		$output = $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-php-versions --raw');
		if(empty($output)) return false;
		
		$output = rtrim(trim($output), ',');
		//print_r($output);
		$lines = explode("\n", $output);
		//print_r($lines);
		foreach($lines as $line){
			$cols = explode(',', $line);
			//print_r($cols);
			if(!preg_match('@^PHP\s+(\d\d?\.\d\d?)$@', $cols[1], $matches)) continue;
			$version = $matches[1];
			//echo $version."\n";
			$this->target_php_versions[$version] = array(
															'major_version' => $version,
															'minor_version' => $cols[5],
															'name' => $cols[1],
															'php_fpm_init_script' => $cols[2],
															'php_fpm_ini_dir' => $cols[3],
															'php_fpm_pool_dir' => $cols[4],
														);
		}
		ksort($this->target_php_versions, SORT_NATURAL);
		//print_r($this->target_php_versions);
	}
	
	// function to get default vhost on the remote server
	public function get_target_default_vhost() {
		global $ssh;
		
		$output = $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-vhosts=domain_id:1 --select=fastcgi_php_version,ssl_key,ssl_request,ssl_cert,ssl_bundle --hide=id,active,Client,Server,"Directive Snippet" --raw');
		if(empty(trim($output))) return false;
		
		//print_r($output);
		$output = rtrim(trim($output), ',');
		$records = explode(',', $output);
		//print_r($records);
        if(count($records) >= 6) {
            $this->web1 = array(
                'domain' => $records[0],
                'fastcgi_php_version' => $records[1],
                'ssl_key' => $records[2],
                'ssl_request' => $records[3],
                'ssl_cert' => $records[4],
                'ssl_bundle' => $records[5],
            );
        }
		//print_r($this->web1);
	}
	
	public function target_is_empty(){
		global $ssh, $argv;
		
		if(isset($argv) && is_array($argv) && in_array('--incremental-db', $argv, true)) {
			return true;
		}
		if(isset($argv) && is_array($argv) && in_array('--data-only', $argv, true)) {
			return true;
		}
		
		$output = $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-vhosts --raw');
		$output = trim($output);
		//if(empty($output)) return true;
		$lines = explode("\n", $output);
		if(count($lines) > 1) return false;
		
		unset($output);
		$output = $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-databases --raw');
		$output = trim($output);
		if(!empty($output)) return false;
		
		return true;
	}
	
	// function to delete default vhost on the remote server
	public function delete_target_default_vhost() {
		global $ssh;

		$output = $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-vhosts=domain_id:1 --raw');
		$output = trim($output);
		if(!empty($output)){
            $ssh->exec("mkdir -p /root/ispcopy/default-vhost-ssl");
            $ssh->exec("mv /var/www/clients/client1/web1/ssl/* /root/ispcopy/default-vhost-ssl/");
			$ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --delete-vhost=id:1');
			$this->swriteln($this->lng('Complete tasks on target server.'));
            $this->runServerSh("target");
			// delete group client1 on target server
			$ssh->exec('groupdel client1');
		}
	}
	
	// do a resync on the remote server
	public function target_resync() {
		global $ssh, $master_ssh;

        if(empty($this->getInputLogEntry("run_resync")) || $this->getInputLogEntry("run_resync") == "y") {
            $this->swriteln($this->lng('Performing a resync on target system...'));
            if ($this->server_role != 'slave') {
                $ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --resync');
            } else {
                $this->swriteln($this->lng('The target server is a slave server. The resync will now be made on the master server.'));
                $master_ssh->exec('/usr/local/ispconfig/server/scripts/console/ISPCConsole.php --resync=server_id:' . ($this->mirror_server_id > 0 ? $this->mirror_server_id : $this->server_id) . ',resync_server:1,resync_server_ip:1,resync_sites:1,resync_ftp:1,resync_shell:1,resync_cron:1,resync_db:1,resync_redis:1,resync_mail:1,resync_mailbox:1,resync_dns:1,resync_f2b_whitelist:1,resync_server_names:1');
            }
        }
	}
	
	// enable/disable interface maintenance mode
	public function set_interface_maintenance($enable) {
		global $ssh, $master_ssh;
		
		// interface maintenance mode
		if($enable){
			$cmd = 'touch '.$this->maintenance_file;
		} else {
			$cmd = 'rm -f '.$this->maintenance_file;
		}
		
		if ($this->server_role != 'slave') {
			$this->swriteln($this->lng('The interface maintenance mode will now be '.($enable? 'enabled' : 'disabled').' on the source and target server.'));
			exec($cmd);
			$ssh->exec($cmd);
		} else {
			$this->swriteln($this->lng('The source and target server are slave servers. The interface maintenance mode will now be '.($enable? 'enabled' : 'disabled').' on the master server.'));
			$master_ssh->exec($cmd);
		}
	}
	
	// enable/disable server maintenance mode
	// $server: both|source|target
	public function set_server_maintenance($enable, $server = 'both') {
		global $ssh;
		
		// server maintenance mode
		if($enable){
			$cmd = 'touch '.$this->server_maintenance_file;
		} else {
			$cmd = 'rm -f '.$this->server_maintenance_file;
		}
		
		$string = 'source and target';
		if($server != 'both'){
			$string = $server;
		}
		$this->swriteln($this->lng('The server maintenance mode will now be '.($enable? 'enabled' : 'disabled').' on the '.$string.' server.'));
		if($server == 'both' || $server == 'source'){
			exec($cmd);
		}
		if($server == 'both' || $server == 'target'){
			$ssh->exec($cmd);
		}
	}
	
	// Migrate the SSH Users
	public function migrate_ssh_users() {
		global $ssh, $app, $conf, $server_config;
		
		// variables
		$commands = array();
		$conflict = false;
		$conflict_msg = '';
		
		$this->log('Migrating SSH users.');
		
		// Read in the passwd and group file of the local system
		$user_array = $this->passwd_file_to_array('/etc/passwd','/etc/shadow');
		$group_array = $this->group_file_to_array('/etc/group');
		
		// Read in the passwd file of the remote system
		$tmpfile = tempnam(sys_get_temp_dir(), "ssh");
		$ssh->getFile('/etc/passwd', $tmpfile);
		$target_user_array = $this->passwd_file_to_array($tmpfile);
		unlink($tmpfile);
		
		// Read in the group file of the remote system
		$tmpfile = tempnam(sys_get_temp_dir(), "ssh");
		$ssh->getFile('/etc/group', $tmpfile);
		$target_group_array = $this->group_file_to_array($tmpfile);
		unlink($tmpfile);
		
		// add the ispconfig and apps user and group
		$ispconfig_system_groups = array('vmail','getmail','sshusers','ispapps','ispconfig');
		$ispconfig_system_users = array('vmail','getmail','ispapps','ispconfig');
		$ispconfig_groups = $ispconfig_system_groups;
		$ispconfig_users = $ispconfig_system_users;
		
		// Loop trough the websites and ssh users in ISPConfig
		$sites = $app->db->queryAllRecords("SELECT domain_id,system_user,system_group FROM web_domain WHERE type = 'vhost' AND server_id = ? ORDER by domain_id ASC", $this->server_id);
		foreach($sites as $site) {
			// Add the website group
			if(!in_array($site['system_group'],$ispconfig_groups)) $ispconfig_groups[] = $site['system_group'];
			
			// add the website user
			if(!in_array($site['system_user'],$ispconfig_users)) $ispconfig_users[] = $site['system_user'];
			
			// add the SSH users of the website
			$records = $app->db->queryAllRecords("SELECT username FROM shell_user WHERE parent_domain_id = ?", (int)$site['domain_id']);
			if(is_array($records)) {
				foreach($records as $rec) {
					if(!in_array($rec['username'],$ispconfig_users)) $ispconfig_users[] = $rec['username'];
				}
			}
		}
		
		// add ISPConfig groups
		foreach($ispconfig_groups as $group) {
			$gid = $group_array[$group]['gid'];
			if((isset($target_group_array[$group]) && $target_group_array[$group]['gid'] != $gid) ||
			   (isset($target_group_array['_'.$gid]) && $target_group_array['_'.$gid] != $group)) {
				// we have a group conflict
				$conflict = true;
				$conflict_msg .= "Group conflict: $group $gid\n";
			}
			// add the group if it does not exist on target or has a different gid
			if(!isset($target_group_array[$group]) || $group_array[$group]['gid'] != $target_group_array[$group]['gid']) {
				// add command to create the group into the queue
				$cmd = 'groupadd --force --non-unique --gid '.$gid.' '.escapeshellarg($group);
				//echo "EXEC: $cmd\n";
				$commands[] = $cmd;
				// add webserver user to group
				if($group != 'vmail' && $group != 'getmail' && $group != 'sshusers') {
					$commands[] = 'usermod -a -G '.$group.' '.$server_config['web']['user'];
				}
			}
		}
		
		// add ISPConfig users
		foreach($ispconfig_users as $user) {
			$uid = $user_array[$user]['uid'];
			if((isset($target_user_array[$user]) && $target_user_array[$user]['uid'] != $uid) ||
			   (isset($target_user_array['_'.$uid]) && $target_user_array['_'.$uid] != $user)) {
				// we have a user conflict
				$conflict = true;
				$conflict_msg .= "User conflict: source($user:$uid) target($user:".$target_user_array[$user]['uid'].")\n";
			}
			// add the user if it does not exist on target or has a different uid
			if(!isset($target_user_array[$user]) || $user_array[$user]['uid'] != $target_user_array[$user]['uid']) {
				// add command to create the user into the queue
				if(in_array($user,$ispconfig_system_users)) {
					$groupname = $user;
					$additionalgroup = '';
				} else {
					$tmp = $app->db->queryOneRecord("SELECT `system_group` FROM `web_domain` WHERE `system_user` = ? AND `system_group` IS NOT NULL AND `type` = 'vhost'", $user);
					if(is_array($tmp) && isset($tmp['system_group'])) {
						$groupname = $tmp['system_group'];
					} else {
						// We have to lookup the system group through shell_user table
						$tmp = $app->db->queryOneRecord("SELECT pgroup FROM shell_user WHERE username = ?", $user);
						$groupname = $tmp['pgroup'];
					}
					if($server_config['web']['add_web_users_to_sshusers_group'] == 'y') {
						$additionalgroup = 'sshusers';
					} else {
						$additionalgroup = '';
					}
				}
				if(!empty($groupname)){
					$gid = $group_array[$groupname]['gid'];
					if(!empty($gid)){
						if($additionalgroup != '') {
							$cmd = 'useradd --non-unique -d '.escapeshellarg($user_array[$user]['home'])." -g $groupname --uid ".$user_array[$user]['uid']." --gid $gid -G $additionalgroup -s ".escapeshellarg($user_array[$user]['shell'])." ".$user;
							//if($apache_chrooted) $app->system->_exec('chroot '.escapeshellcmd($web_config['website_basedir']).' useradd -d '.escapeshellcmd($data['new']['document_root'])." -g $groupname $fixed_uid_param -G sshusers $username -s /bin/false");
						} else {
							$cmd = 'useradd --non-unique -d '.escapeshellarg($user_array[$user]['home'])." -g $groupname --uid ".$user_array[$user]['uid']." --gid $gid -s ".escapeshellarg($user_array[$user]['shell'])." ".$user;
							//if($apache_chrooted) $app->system->_exec('chroot '.escapeshellcmd($web_config['website_basedir']).' useradd -d '.escapeshellcmd($data['new']['document_root'])." -g $groupname $fixed_uid_param $username -s /bin/false");
						}
					}
				}
				
				// To work around a limitation in the useradd command in CentOS,
				// we will have  to create the base path first
				$path_parts = explode('/',$user_array[$user]['home']);
				unset($path_parts[count($path_parts)-1]);
				$base_path = implode('/',$path_parts);
                $mkdir_cmd = '//##mkdir##//';
                if(strpos($base_path, '/./') != false) {
                    $lsattr = "lsattr -d " . explode('/./', $base_path)[0] . " | grep -- '-i-'";
                    exec($lsattr, $check_immutable_out);
                    if(!empty($check_immutable_out)) {
                        $mkdir_cmd = 'mkdir -p ' . explode('/./', $base_path)[0] . ';chattr -i ' . explode('/./', $base_path)[0] . ';' . $mkdir_cmd . ';chattr +i ' . explode('/./', $base_path)[0] . ';';
                    }
                }
				$base_path = str_replace('/./','/',$base_path);
				if(!$ssh->isDir($base_path)) {
                    $commands[] = str_replace('//##mkdir##//', 'mkdir -p ' . escapeshellarg($base_path), $mkdir_cmd);
                }
				
				//echo "EXEC: $cmd\n";
				$commands[] = $cmd;
			}
		}
		
		// user and group conflict management
		if($conflict) {
			$this->swrite($conflict_msg);
			$this->log('User or group conflict:'.$conflict_msg,1);
			$result = $this->simple_query('Proceed anyway?', array('y','n'), 'n', 'conflict_proceed');
			if($result == 'y') {
				$conflict = false;
				$this->log('Conflict: Proceed.');
			}
		}
		
		if($conflict) {
			$this->log('Conflict: Stopped.');
			// Wartungsmodus deaktivieren
			$this->set_maintenance(false);
			die("Program stopped.\n");
		} else {
			// Execute commands on remote server
			//die(print_r($commands));
			foreach($commands as $cmd) {
				if($ssh->exec($cmd) === false) {
					$this->warning('Remote command failed: '.$cmd."\n");
				}
			}
		}
		
	}
	
	// function to move a mysql user to the remote server
	private function migrate_mysql_user($username, $force_sync = false) {
		global $ssh, $argv;
		
		if(isset($argv) && is_array($argv) && in_array('--exclude-db', $argv, true)) {
				return true;
		}

		$current_grant_hash = false;
		if(!$force_sync && isset($argv) && is_array($argv) && in_array('--incremental-db', $argv, true)) {
			if(file_exists(BASE_PATH . '/data/' . $username . '.grants')) {
				$tmp = '';
				$fp = gzopen(BASE_PATH . '/data/' . $username . '.grants', 'r');
				while($fp && !gzeof($fp)) {
					$tmp .= gzgets($fp);
				}
				gzclose($fp);
				
				$tmp = trim($tmp);
				if($tmp) $current_grant_hash = $tmp;
			}
		}

		$this->log('Migrating MySQL user: '.$username);
		$this->swriteln($this->lng('Migrating MySQL User:').' '.$username);
		
		// TODO: Check if user exists
		// TODO: preg check for username, just to be sure
		
		$new_grant_hash = false;
		
		// dump the user permissions to a file
		$tmpdir = sys_get_temp_dir().'/'.uniqid();
		mkdir($tmpdir);
		$mysql_dump_file = $tmpdir.'/user_privileges_final.sql';
		$mysql_dump_file_remote = sys_get_temp_dir().'/user_privileges_final.sql';
		
		// Get MySQL version
		exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -e\"SELECT VERSION();\" > $tmpdir/mysql_version.txt");
		$tmp = explode("\n",file_get_contents($tmpdir.'/mysql_version.txt'));
		$tmp2 = explode('-',$tmp[1]);
		$mysql_version = $tmp2[0];
		unlink($tmpdir.'/mysql_version.txt');
		$this->log('MySQL version: '.$mysql_version);
		
		// If MySQL > 5.7 but not MariaDB
		if(version_compare($mysql_version, '5.7.0', '>=') && version_compare($mysql_version, '10.0.0','<=')) {
			exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -e\"select concat('show create user ','\'',user,'\'@\'',host,'\'') from mysql.user WHERE mysql.user.User ='$username'\" > $tmpdir/user_create_list_with_header.txt");
			if(file_exists($tmpdir.'/user_create_list_with_header.txt') && filesize($tmpdir.'/user_create_list_with_header.txt') > 0) {
				exec("sed '1d' $tmpdir/user_create_list_with_header.txt > $tmpdir/user_create.txt");
				exec("while read user_create; do mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -e".'"$user_create"'." > $tmpdir/user_create_grant.txt; sed '1d' $tmpdir/user_create_grant.txt >> $tmpdir/user_create_privileges.txt; done < $tmpdir/user_create.txt");
				exec("awk '{print $0\";\"}'  $tmpdir/user_create_privileges.txt > $mysql_dump_file");
				exec("rm $tmpdir/user_create.txt $tmpdir/user_create_list_with_header.txt $tmpdir/user_create_grant.txt $tmpdir/user_create_privileges.txt");
			}
		}
		
		exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -e\"select concat('show grants for ','\'',user,'\'@\'',host,'\'') from mysql.user WHERE mysql.user.User ='$username'\" > $tmpdir/user_list_with_header.txt");
		if(file_exists($tmpdir.'/user_list_with_header.txt') && filesize($tmpdir.'/user_list_with_header.txt') > 0) {
			exec("sed '1d' $tmpdir/user_list_with_header.txt > $tmpdir/user.txt");
			
			$new_grant_hash = sha1(file_get_contents($tmpdir . '/user.txt'));
			if(!$current_grant_hash || $current_grant_hash != $new_grant_hash) {
				exec("while read user; do mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -e".'"$user"'." > $tmpdir/user_grant.txt; sed '1d' $tmpdir/user_grant.txt >> $tmpdir/user_privileges.txt; echo \"flush privileges\" >> $tmpdir/user_privileges.txt; done < $tmpdir/user.txt");
				exec("awk '{print $0\";\"}'  $tmpdir/user_privileges.txt >> $mysql_dump_file");
				exec("rm $tmpdir/user.txt $tmpdir/user_list_with_header.txt $tmpdir/user_grant.txt $tmpdir/user_privileges.txt");

                // escape column names
                $mysql_dump_content = file_get_contents($mysql_dump_file);
                if (preg_match_all("@\(.*\)@", $mysql_dump_content, $matches)) {
                    foreach ($matches[0] as $match) {
                        $mysql_dump_content = str_replace($match, preg_replace("@\b@", '`', $match), $mysql_dump_content);
                    }
                    file_put_contents($mysql_dump_file, $mysql_dump_content);
                }

                // Copy the dump file to the target
                $ssh->putFile($mysql_dump_file,$mysql_dump_file_remote);
				
				// Import the dump file
				if($ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' < '.$mysql_dump_file_remote) === false) {
					// Keep error file on remote server
					$ssh->mv($mysql_dump_file_remote,$mysql_dump_file_remote.'.'.$username.'.err');
				} else {
					// Remote cleanup
					$ssh->exec('rm -f '.escapeshellarg($mysql_dump_file_remote));
				}
				
				// local cleanup
				unlink($mysql_dump_file);
				rmdir($tmpdir);
			}
		} else {
			unlink("$tmpdir/user_list_with_header.txt");
			rmdir($tmpdir);
		}
	
		if(!is_dir(BASE_PATH . '/data')) mkdir(BASE_PATH . '/data', 0777);
		$fp = gzopen(BASE_PATH . '/data/' . $username . '.grants', 'w');
		gzwrite($fp, $new_grant_hash);
		gzclose($fp);
	
	}
	
	// function to move a mysql db to the remote server
	public function migrate_mysql_db($db_name, $force_full = false) {
		global $ssh, $argv, $app, $conf;

        $app->uses("dbmaster");

		// get the local MySQL root password
		/*
		if(!is_file('/usr/local/ispconfig/server/lib/mysql_clientdb.conf')) $this->error('Local file /usr/local/ispconfig/server/lib/mysql_clientdb.conf not found.');
		include '/usr/local/ispconfig/server/lib/mysql_clientdb.conf';
		
		$this->local_mysql_root_user = $clientdb_user;
		$this->local_mysql_root_password = $clientdb_password;
		$this->local_mysql_host = $clientdb_host;
		*/
		
		if(isset($argv) && is_array($argv) && in_array('--exclude-db', $argv, true)) {
			return true;
		}	
		if(isset($argv) && is_array($argv) && in_array('--exclude-db=' . $db_name, $argv, true)) {
			$this->log('Skipping MySQL DB: ' . $db_name . ' as requested.');
			return true;
		}
		$table_hashes = array();
		if(!$force_full && isset($argv) && is_array($argv) && in_array('--incremental-db', $argv, true)) {
			if(file_exists(BASE_PATH . '/data/' . $db_name . '.dat')) {
				$tmp = '';
				$fp = gzopen(BASE_PATH . '/data/' . $db_name . '.dat', 'r');
				while($fp && !gzeof($fp)) {
					$tmp .= gzgets($fp);
				}
				gzclose($fp);
				
				$tmp = unserialize($tmp);
				if($tmp && is_array($tmp)) $table_hashes = $tmp;
			}
		}
		
		$this->log('Migrating MySQL DB: '.$db_name);
		$this->swriteln($this->lng('Migrating MySQL DB:').' '.$db_name);
		
		// TODO: Check if db exists
		// TODO: preg check for db names, just to be sure
		
		
		$new_table_hashes = array();
		$diff_tables = array();
		$delete_tables = array();
		
		// get table status hashes
		$tmp = shell_exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -s -N -D ' . escapeshellarg($db_name) . ' -e "SHOW TABLE STATUS"');
		$lines = explode("\n", trim($tmp));
		for($i = 0; $i < count($lines); $i++) {
			$line = trim($lines[$i]);
			$cols = preg_split('/\t+/', $line, 2);
			$table_name = $cols[0];
			$table_hash = sha1($line);
			$new_table_hashes[$table_name] = $table_hash;
			if(!array_key_exists($table_name, $table_hashes) || $table_hashes[$table_name] != $table_hash) $diff_tables[] = $table_name;
		}
		
		// check for tables that were deleted
		reset($table_hashes);
		foreach($table_hashes as $table_name => $table_hash) {
			if(!array_key_exists($table_name, $new_table_hashes)) $delete_tables[] = $table_name;
		}
		
		if(count($delete_tables) > 0) {
			for($i = 0; $i < count($delete_tables); $i++) {
				$ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "DROP TABLE ' . escapeshellarg($delete_tables[$i]) . '"');
			}
		}
		
		if(count($diff_tables) > 0) {
			// get source database charset
			$tmp = exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -s -N -D ' . escapeshellarg($db_name) . ' -e "SHOW VARIABLES LIKE \'character_set_database\'"');
			if($tmp) list($dummy, $charset) = preg_split('/\s+/', trim($tmp));
			if(empty($charset)) $charset = 'utf8';
			
			// Create the database on the remote system, if it does not exist
			$ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -e "CREATE DATABASE IF NOT EXISTS '.$db_name.' DEFAULT CHARSET ' . escapeshellarg($charset) . ';"');
			
			// dump the local database
			$limit_table_arg = '';
			if(count($table_hashes) > 0) {
				for($i = 0; $i < count($diff_tables); $i++) {
					$limit_table_arg .= ' ' . escapeshellarg($diff_tables[$i]);
				}
			}
			
			$mysql_dump_file = tempnam(sys_get_temp_dir(), "mysql-");
			//echo $mysql_dump_file."\n";
			if($db_name != $conf['db_database']){
                $cmd =$this->copyUserDB($db_name, $limit_table_arg, $mysql_dump_file);
			} else {
                $cmd = $this->copyDbIspconfig($db_name, $diff_tables, $mysql_dump_file);
                chmod($mysql_dump_file,0600);

                // Copy the dump file to the target
                $ssh->putFile($mysql_dump_file,$mysql_dump_file);

                // Import the dump file
                $ssh->exec('zcat ' . escapeshellarg($mysql_dump_file) . ' | mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -f --max_allowed_packet=512M -D '.escapeshellarg($db_name));

                // cleanup
                $ssh->exec('rm -f '.escapeshellarg($mysql_dump_file));
                unlink($mysql_dump_file);
			}

            $this->log('Executed local command '.$cmd,0);
		}
		
		if(!is_dir(BASE_PATH . '/data')) mkdir(BASE_PATH . '/data', 0777);
		$fp = gzopen(BASE_PATH . '/data/' . $db_name . '.dat', 'w');
		gzwrite($fp, serialize($new_table_hashes));
		gzclose($fp);
	}

    private function copyUserDB($db_name, $limit_table_arg, $mysql_dump_file) {
        global $ssh;
        // --single-transaction (InnoDB only) or not?
        $single_transaction = '';
        $link = new mysqli($this->local_mysql_host, $this->local_mysql_root_user, $this->local_mysql_root_password);
        $res = $link->query("SHOW TABLE STATUS FROM `" . $db_name . "` WHERE Engine != 'InnoDB'");

        if (isset($res) && $res !== false) {
            $aResults = array();
            while ($row = $res->fetch_assoc()) {
                $aResults[] = $row;
                // one result row is enough in this case
                if (!empty($aResults)) break;
            }
            $res->free();
            if (is_array($aResults) && empty($aResults)) $single_transaction = ' --single-transaction';
            unset($aResults);
        }
        $link->close();

//        $cmd = 'mysqldump --defaults-extra-file='.$this->mysql_defaults_extra_file.' -cCeQ -f --hex-blob --quote-names --quick'.$single_transaction.' --routines --events --triggers '.escapeshellarg($db_name) . $limit_table_arg . ' | gzip -6 > '.escapeshellarg($mysql_dump_file);
        $cmd = 'mysqldump --defaults-extra-file='.$this->mysql_defaults_extra_file.' -cCeQ -f --hex-blob --quote-names --quick'.$single_transaction.' --routines --events --max_allowed_packet=512M --triggers '.escapeshellarg($db_name) . $limit_table_arg;
        $cmd .= ' | ssh -i ' . $ssh->getIdentityFile() . " root@" . $this->target_server_hostname . ' "mysql --defaults-extra-file=' . $this->mysql_defaults_extra_file . ' -f --max_allowed_packet=512M -D ' . escapeshellarg($db_name). '"';
        exec($cmd);
        return $cmd;
    }

    private function copyDbIspconfig($db_name, $diff_tables, $mysql_dump_file) {
        global $app, $ssh;

        $mysql_dump_file2 = tempnam(sys_get_temp_dir(), "mysql-");
        $server = $app->db->queryOneRecord("SELECT * FROM server WHERE server_id = ?", $this->server_id);
        $run_server_sh = false;
        // loop through ISPConfig tables
        foreach($diff_tables as $diff_table){
            $where = '';  // kann benutzt werden, um beim mysqldump bestimmte Records auszuschliessen; in solch einem Fall wird die Zieltabelle auf dem Zielserver NICHT gelöscht und nicht neu erstellt, und die Records vom mysqldump werden zur existierenden Zieltabelle hinzugefügt
            $where2 = ''; // kann benutzt werden, um beim mysqldump bestimmte Records auszuschliessen; die Zieltabelle auf dem Zielserver wird gelöscht und neu erstellt und mit den Records aus dem mysqldump befüllt
            $restore_web1 = false;

            switch ($diff_table) {
                case 'attempts_login':
                    //case 'f2b_current_bans':
                    //case 'f2b_whitelist':
                    //case 'mail_backup':
                    //case 'server':
                    //case 'server_ip':
                    //case 'server_php':
                case 'sys_config':
                case 'sys_cron':
                    //case 'sys_datalog':
                    //case 'sys_log':
                    //case 'sys_remoteaction':
                case 'sys_session':
                    //case 'web_backup':
                    continue 2;
                    break;
                case 'f2b_current_bans':
                    // server_id anpassen
                    //if($this->server_id != 1) $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE f2b_current_bans SET server_id = '.$this->server_id.' WHERE server_id = 1"');

                    // auf Slaves ist die f2b_current_bans-Tabelle immer leer
                    if($this->server_role == 'slave'){
                        $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "DELETE FROM f2b_current_bans WHERE 1"');
                    }
                    // ansonsten baut sich die Tabelle von selber auf bzw. wird von selbst angepasst
                    continue 2;
                    break;
                /*
                case 'f2b_whitelist':
                    //if($this->server_role == 'single') continue 2;

                    // Single, Master & Slave: Tabelle übertragen und Resync auslösen
                    break;
                */
                case 'mail_backup':
                case 'web_backup':
                    // Tabelle auf Zielserver leeren
                    $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "DELETE FROM '.$diff_table.' WHERE 1"');

                    if($this->server_role == 'single'){
                        // auf Single-Servern müssen die Backups neu erstellt werden
                        continue 2;
                    }

                    if($this->server_role == 'slave'){
                        // auf Slave-Servern müssen die Backups neu erstellt werden und die alten Backups aus der mail_backup/web_backup-Tabelle auf dem Master gelöscht werden
                        $app->dbmaster->query("DELETE FROM ?? WHERE `server_id` = ?", $diff_table, $this->server_id); // für Production wieder aktivieren
                        continue 2;
                    }

                    // Master: Backups der anderen Server übernehmen, eigene Backups müssen neu erstellt werden
                    $where = '--where="server_id != '.$this->server_id.'"';
                    break;
                case 'server':
                    if($this->server_role == 'single'){
                        $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.', updated = '.$server['updated'].' WHERE server_id = 1"');
                        // Autoincrement für sys_datalog setzen
                        $sys_datalog_auto_increment = "ALTER TABLE sys_datalog AUTO_INCREMENT = ".($server['updated'] + 1000).";";

                        // server-Tabelle nicht auf Zielserver kopieren
                        continue 2;
                    }

                    $where = '--where="server_id != '.$this->server_id.'"';

                    if($this->server_role == 'master'){
                        $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.', updated = '.$server['updated'].' WHERE server_id = 1"');
                        // Autoincrement für sys_datalog setzen
                        $sys_datalog_auto_increment = "ALTER TABLE sys_datalog AUTO_INCREMENT = ".($server['updated'] + 1000).";";
                    } else {
                        // slave
                        $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.', mail_server = '.$server['mail_server'].', web_server = '.$server['web_server'].', cloud_server = '.$server['cloud_server'].', loadbalancer = '.$server['loadbalancer'].', backend_web_server = '.$server['backend_web_server'].', dns_server = '.$server['dns_server'].', file_server = '.$server['file_server'].', db_server = '.$server['db_server'].', vserver_server = '.$server['vserver_server'].', proxy_server = '.$server['proxy_server'].', firewall_server = '.$server['firewall_server'].', mirror_server_id = '.$server['mirror_server_id'].', updated = '.$server['updated'].' WHERE server_id = 1"');
                    }
                    break;
                case 'server_ip':
                    if($this->server_role == 'single') continue 2;
                    if($this->server_role == 'master'){
                        // get max server_ip_id value from master
                        $max_server_ip_id = $app->db->queryOneRecord("SELECT MAX(server_ip_id) AS max_server_ip_id FROM server_ip")['max_server_ip_id'];
                        $max_server_ip_id = intval($max_server_ip_id);
                        if($max_server_ip_id > 0) $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_ip_id = server_ip_id + '.$max_server_ip_id.' WHERE server_ip_id > 1"');

                        $where = '--where="server_id != '.$this->server_id.'"';
                        if($this->server_id != 1) $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.' WHERE server_id = 1"');
                    }
                    if($this->server_role == 'slave'){
                        // server_ip-Tabelle wird übernommen, IPs werden später auf dem Target-Server in der Funktion replace_ip() angepasst sowie auf dem Master-Server in der Funktion master_change_mysql_access()
                    }
                    break;
                case 'server_php':
                    if($this->server_role == 'single') continue 2;

                    if($this->server_role == 'master' || $this->server_role == 'slave'){
                        //$ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.' WHERE server_id = 1"');

                        $server_php_ids_to_use_on_target_server = array();
                        if(!empty($this->target_php_versions)){
                            foreach($this->target_php_versions as $target_php_version){
                                $tmp_php_version = $app->db->queryOneRecord("SELECT * FROM ?? WHERE php_fpm_init_script = ? AND server_id = ?", $diff_table, $target_php_version['php_fpm_init_script'], $this->server_id);
                                if(!empty($tmp_php_version) && is_array($tmp_php_version)){
                                    if(!in_array($tmp_php_version['server_php_id'], $server_php_ids_to_use_on_target_server)) $server_php_ids_to_use_on_target_server[] = $tmp_php_version['server_php_id'];
                                }
                                unset($tmp_php_version);
                            }
                        }
                        $php_versions_from_other_servers = $app->db->queryAllRecords("SELECT * FROM ?? WHERE server_id != ?", $diff_table, $this->server_id);
                        // remove duplicate PHP versions from array
                        $php_versions_by_server_id = array();
                        $php_version_count = count($php_versions_from_other_servers);
                        for($i=0;$i<$php_version_count;$i++){
                            if(!empty($php_versions_by_server_id[$php_versions_from_other_servers[$i]['server_id']]) && is_array($php_versions_by_server_id[$php_versions_from_other_servers[$i]['server_id']]) && in_array($php_versions_from_other_servers[$i]['php_fpm_init_script'], $php_versions_by_server_id[$php_versions_from_other_servers[$i]['server_id']])) {
                                unset($php_versions_from_other_servers[$i]);
                            } else {
                                $php_versions_by_server_id[$php_versions_from_other_servers[$i]['server_id']][] = $php_versions_from_other_servers[$i]['php_fpm_init_script'];
                            }
                        }
                        unset($php_versions_by_server_id, $php_version_count);

                        if(!empty($php_versions_from_other_servers) && is_array($php_versions_from_other_servers)){
                            foreach($php_versions_from_other_servers as $php_v){
                                if(!in_array($php_v['server_php_id'], $server_php_ids_to_use_on_target_server)) $server_php_ids_to_use_on_target_server[] = $php_v['server_php_id'];
                            }
                        }

                        if(!empty($server_php_ids_to_use_on_target_server)){
                            $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "DELETE FROM '.$diff_table.' WHERE 1"');

                            $where = '--where="server_php_id IN('.implode(',', $server_php_ids_to_use_on_target_server).')"';
                        }
                    }
                    break;
                case 'sys_datalog':
                    $where2 = '--where="dbtable NOT IN(\'instant_backup\',\'server_ip\')"';
                case 'sys_log':
                case 'sys_remoteaction':
                    // server_id anpassen
                    if($this->server_id != 1) $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($db_name) . ' -e "UPDATE '.$diff_table.' SET server_id = '.$this->server_id.' WHERE server_id = 1"');
                    break;
                /*
                case 'client':
                    $where = '--where="client_id > 1"';
                    break;
                */
                case 'web_domain':
                    //$where = '--where="domain_id > 1"';
                    //print_r($this->web1);
                    if(!empty($this->web1) && is_array($this->web1)){
                        $web1 = $app->db->queryOneRecord("SELECT * FROM web_domain WHERE domain_id = 1");
                        if(!empty($web1) && is_array($web1)){
                            if($web1['domain'] == $server['server_name'] && $web1['domain'] != $this->web1['domain']){
                                $restore_web1 = true;
                                $app->db->query("UPDATE web_domain SET domain = ?, fastcgi_php_version = ?, ssl_key = ?, ssl_request = ?, ssl_cert = ?, ssl_bundle = ? WHERE domain_id = 1", $this->web1['domain'], $this->web1['fastcgi_php_version'], $this->web1['ssl_key'], $this->web1['ssl_request'], $this->web1['ssl_cert'], $this->web1['ssl_bundle']);
                            }
                        }
                    }

                    // map PHP versions
                    if($restore_web1){
                        $webs = $app->db->queryAllRecords("SELECT * FROM web_domain WHERE domain_id > 1");
                    } else {
                        $webs = $app->db->queryAllRecords("SELECT * FROM web_domain");
                    }
                    if(!empty($webs) && is_array($webs)){
                        foreach($webs as $web){
                            $web['fastcgi_php_version'] = trim($web['fastcgi_php_version']);
                            if(empty($web['fastcgi_php_version'])) continue;
                            if(!preg_match('@(\d\d?\.\d\d?)@',$web['fastcgi_php_version'],$matches)) continue;
                            $major_version = $matches[1];
                            if(isset($this->target_php_versions[$major_version])){
                                // PHP 5.3.27:/etc/init.d/php-5.3.27-fpm:/opt/php-5.3.27/lib:/opt/php-5.3.27/etc/pool.d
                                $fastcgi_php_version = $this->target_php_versions[$major_version]['name'].':'.$this->target_php_versions[$major_version]['php_fpm_init_script'].':'.$this->target_php_versions[$major_version]['php_fpm_ini_dir'].':'.$this->target_php_versions[$major_version]['php_fpm_pool_dir'];
                            } else {
                                foreach($this->target_php_versions as $key => $val){
                                    if(version_compare($key, $major_version, '>')){
                                        $fastcgi_php_version = $this->target_php_versions[$key]['name'].':'.$this->target_php_versions[$key]['php_fpm_init_script'].':'.$this->target_php_versions[$key]['php_fpm_ini_dir'].':'.$this->target_php_versions[$key]['php_fpm_pool_dir'];
                                        break;
                                    }
                                }
                            }
                            if(isset($fastcgi_php_version)){
                                $app->db->query("UPDATE web_domain SET fastcgi_php_version = ? WHERE domain_id = ?", $fastcgi_php_version, $web['domain_id']);
                            }
                            unset($fastcgi_php_version);
                        }
                    }
                    break;
            }
            if(!empty($where)){
                $where .= ' --skip-add-drop-table --no-create-info ';
                $where2 = '';
            }
            if(!empty($where2)) $where2 = trim($where2).' ';

            $cmd = 'mysqldump --defaults-extra-file='.$this->mysql_defaults_extra_file.' -cCeQ -f --hex-blob --quote-names --quick --routines --events --max_allowed_packet=512M --triggers '.$where.$where2.escapeshellarg($db_name) .' '. escapeshellarg($diff_table) . ' >> '.escapeshellarg($mysql_dump_file2);
            //echo $cmd."\n";
            exec($cmd);

            if($restore_web1){
                $app->db->query("UPDATE web_domain SET domain = ?, fastcgi_php_version = ?, ssl_key = ?, ssl_request = ?, ssl_cert = ?, ssl_bundle = ? WHERE domain_id = 1", $web1['domain'], $web1['fastcgi_php_version'], $web1['ssl_key'], $web1['ssl_request'], $web1['ssl_cert'], $web1['ssl_bundle']);
            }
            // restore original PHP versions
            if(!empty($webs) && is_array($webs)){
                foreach($webs as $web){
                    $app->db->query("UPDATE web_domain SET fastcgi_php_version = ? WHERE domain_id = ?", $web['fastcgi_php_version'], $web['domain_id']);
                }
                unset($webs);
            }
        }
        if(!empty($sys_datalog_auto_increment)){
            file_put_contents($mysql_dump_file2, $sys_datalog_auto_increment, FILE_APPEND);
        }
        exec('cat '.escapeshellarg($mysql_dump_file2).' | gzip -6 > '.escapeshellarg($mysql_dump_file));
        unlink($mysql_dump_file2);

        return $cmd;
    }
	
	// Function to migrate the MySQL databases
	public function migrate_databases() {
		global $ssh, $app, $conf, $argv, $master_ssh;
		
		// get the local MySQL root password
		if(!is_file('/usr/local/ispconfig/server/lib/mysql_clientdb.conf')) $this->error('Local file /usr/local/ispconfig/server/lib/mysql_clientdb.conf not found.');
		include '/usr/local/ispconfig/server/lib/mysql_clientdb.conf';
		
		$this->local_mysql_root_user = $clientdb_user;
		$this->local_mysql_root_password = $clientdb_password;
		$this->local_mysql_host = $clientdb_host;

		// Test the local MySQL password
		$n=0;
		while(!@mysqli_connect($this->local_mysql_host, $this->local_mysql_root_user, $this->local_mysql_root_password)) {
			if($n > 0) $this->swriteln($this->lng('MySQL login failed.'));
			$this->swriteln($this->lng('Enter the login details of the local MySQL root user or the word "quit" to stop the program.'));
			$this->local_mysql_host = $this->free_query('Local MySQL server hostname', 'localhost','mysql_hostname');
			$this->local_mysql_root_user = $this->free_query('Local MySQL root username', 'root','mysql_root_user');
			$this->local_mysql_root_password = $this->free_query('Local MySQL root password', '','mysql_root_password');
			$n++;
		}
		
		// get the remote MySQL root password
		if($ssh->isFile('/usr/local/ispconfig/server/lib/mysql_clientdb.conf')) {
			$tmpfile = tempnam(sys_get_temp_dir(), "mysql-");
			$ssh->getFile('/usr/local/ispconfig/server/lib/mysql_clientdb.conf', $tmpfile);
			include $tmpfile;
			unlink($tmpfile);
		
			$this->remote_mysql_root_user = $clientdb_user;
			$this->remote_mysql_root_password = $clientdb_password;
			$this->remote_mysql_host = $clientdb_host;
		}
		
		// Test the remote user
		$n=0;
		while(!$ssh->testDBLogin($this->remote_mysql_host,$this->remote_mysql_root_user,$this->remote_mysql_root_password)) {
			if($n > 0) $this->swriteln($this->lng('MySQL login failed.'));
			$this->swriteln($this->lng('Enter the login details of the MySQL root user of the target server or the word "quit" to stop the program.'));
			$this->remote_mysql_host = 'localhost';
			$this->remote_mysql_root_user = $this->free_query('Target MySQL root username', 'root','remote_mysql_root_user');
			$this->remote_mysql_root_password = $this->free_query('Target MySQL root password', '','remote_mysql_root_password');
			$n++;
		}
		
		// get the master MySQL root password
		if(!empty($this->master_server)){
			if($master_ssh->isFile('/usr/local/ispconfig/server/lib/mysql_clientdb.conf')) {
				$tmpfile = tempnam(sys_get_temp_dir(), "mysql-");
				$master_ssh->getFile('/usr/local/ispconfig/server/lib/mysql_clientdb.conf', $tmpfile);
				include $tmpfile;
				unlink($tmpfile);
			
				$this->master_mysql_root_user = $clientdb_user;
				$this->master_mysql_root_password = $clientdb_password;
				$this->master_mysql_host = $clientdb_host;
			}
			
			// Test the remote user
			$n=0;
			while(!$master_ssh->testDBLogin($this->master_mysql_host,$this->master_mysql_root_user,$this->master_mysql_root_password)) {
				if($n > 0) $this->swriteln($this->lng('MySQL login failed.'));
				$this->swriteln($this->lng('Enter the login details of the MySQL root user of the master server or the word "quit" to stop the program.'));
				$this->master_mysql_host = 'localhost';
				$this->master_mysql_root_user = $this->free_query('Master MySQL root username', 'root','master_mysql_root_user');
				$this->master_mysql_root_password = $this->free_query('Master MySQL root password', '','master_mysql_root_password');
				$n++;
			}
		}
		
		// Migrate the ISPConfig MySQL user and database
		if(isset($argv) && is_array($argv) && in_array('--data-only', $argv, true)) {
		} else {
			$this->migrate_mysql_db($conf['db_database'], true);
		}
		//$this->migrate_mysql_user($conf['db_user'], true);
		
		// Migrate website databases
		$databases = $app->db->queryAllRecords("SELECT database_name FROM web_database WHERE server_id = ? ORDER BY database_name", $this->server_id);
		if(is_array($databases)) {
			foreach($databases as $db) {
				$this->migrate_mysql_db($db['database_name']);
			}
		}
		
		// migrate website database users
		$dbusers = $app->db->queryAllRecords("SELECT database_user FROM web_database_user WHERE server_id IN ('0,".$this->server_id."') ORDER BY database_user");
		if(is_array($dbusers)) {
			foreach($dbusers as $user) {
				$this->migrate_mysql_user($user['database_user']);
			}
		}
		
		if(isset($argv) && is_array($argv)) {
			for($i = 0; $i < count($argv); $i++) {
				$arg = $argv[$i];
				if(strpos($arg, '=') === false) {
					continue;
				} elseif(substr($arg, 0, 2) !== '--') {
					continue;
				}
				list($key, $val) = explode('=', substr($arg, 2), 2);
				if($key === 'migrate-db') {
					if(strpos($val, ',') !== false) {
						$entries = explode(',', $val);
						for($e = 0; $e < count($entries); $e++) {
							$this->migrate_mysql_db($entries[$e]);
						}
					} else {
						$this->migrate_mysql_db($val);
					}
				} elseif($key == 'migrate-db-user') {
					if(strpos($val, ',') !== false) {
						$entries = explode(',', $val);
						for($e = 0; $e < count($entries); $e++) {
							$this->migrate_mysql_user($entries[$e]);
						}
					} else {
						$this->migrate_mysql_user($val);
					}
				}
			}
		}
	}
	
	// Function to migrate the MySQL databases
	public function migrate_fstab() {
		global $ssh, $app, $conf;
		
		$replace_fstab_line_for_default_vhost = false;
		if(!empty($this->web1) && is_array($this->web1)){
			$web1 = $app->db->queryOneRecord("SELECT * FROM web_domain WHERE domain_id = 1");
			if(!empty($web1) && is_array($web1)){
				$server = $app->db->queryOneRecord("SELECT * FROM server WHERE server_id = ?", $this->server_id);
				if($web1['domain'] == $server['server_name'] && $web1['domain'] != $this->web1['domain']){
					$replace_fstab_line_for_default_vhost = true;
				}
			}
		}
		
		// get remote fstab file
		$tmpfile = tempnam(sys_get_temp_dir(), "fstab-");
		chmod($tmpfile, 0644);
		$ssh->getFile('/etc/fstab', $tmpfile);
		
		// extract bind mount lines from local fstab file
		$fstab = file('/etc/fstab',FILE_IGNORE_NEW_LINES);
		if(is_array($fstab)) {
			foreach($fstab as $line) {
				if($replace_fstab_line_for_default_vhost){
					$line = preg_replace('~^(/var/log/ispconfig/httpd/)('.$web1['domain'].')(\s+/var/www/clients/client)~','${1}'.$this->web1['domain'].'${3}', $line);
				}
				if(substr($line,0,24) == '/var/log/ispconfig/httpd') {
					$app->system->replaceLine($tmpfile, $line, $line, 0, 1);
				}
			}
		}
		
		// Copy fstab file back after doing a backup
		$ssh->cp('/etc/fstab','/etc/fstab~');
		$ssh->putFile($tmpfile,'/etc/fstab');
		
		// activate mounts on target
		$ssh->exec('mount -a');
		
		// cleanup
		unlink($tmpfile);
	}
	
	// Function to migrate the MySQL databases
	public function migrate_files() {
		global $ssh, $server_config, $app, $conf;
		
		// Migrate various config files
		include 'filelist.inc.php';
		
		// Ask if the backup dir shall be migrated
		/*
		$result = $this->simple_query('Shall I copy the ISPConfig backup directory to the new server?', array('y','n'), 'n', 'copy_backup_dir');
		if($result == 'y') {
			$files[] = $server_config['server']['backup_dir'];
		}
		*/
		
		// Copy custom paths when defined
		if(file_exists(BASE_PATH.'/custom-copy-paths.txt') && fileowner(BASE_PATH.'/custom-copy-paths.txt') === 0) {
			$this->log('Loading custom copy paths from: '.BASE_PATH.'/custom-copy-paths.txt');
			$custom_paths = file(BASE_PATH.'/custom-copy-paths.txt',FILE_IGNORE_NEW_LINES);
			if(is_array($custom_paths)) {
				foreach($custom_paths as $p) {
					$files[] = $p;
				}
			}
		}

		// Migrate paths from server config
		foreach($files as $f) {
			if(is_dir($f) || is_link($f)) {
                if($this->getInputLogEntry("mode") == "second" && $f == $server_config['web']['website_basedir']) {
                    $tmp_path = preg_replace("~\[client_id\]~",
                        "*",
                        preg_replace("~/[\w\-]+\[website_id\]~",
                            "",
                            $server_config['web']['website_path']));
                    preg_match("~.*/$~", $tmp_path, $end_slash_check);
                    if(empty($end_slash_check)) {
                        $tmp_path .= "/";
                    }
                    $immutables = $ssh->exec('lsattr ' . $tmp_path . ' 2>/dev/null | awk /-i-/\'{print$2}\'');
                } else if(in_array($this->getInputLogEntry("mode"), ["first", "single"])) {
                    $immutables = $ssh->exec('lsattr -R ' . $f . ' 2>/dev/null | awk /-i-/\'{print$2}\'');
                    $immutables = trim(trim($immutables) . "\n" . $ssh->exec('lsattr -d' . $f . ' 2>/dev/null | awk /-i-/\'{print$2}\''));
                }
                if(!empty($immutables)) {
                    foreach (explode("\n", $immutables) as $d) {
                        $ssh->exec("chattr -i " . $d);
                    }
                }
				$this->swriteln('Copy Directory: '.$f);
				$this->log('Copy Directory: '.$f);
				if(substr($f,-1) != '/') $f .= '/';
				if($f == '/etc/postfix/') $ssh->mv('/etc/postfix/dynamicmaps.cf', '/tmp/dynamicmaps.cf.migsave');
				$ssh->copyFiles($f, $f, 'rsync', true, ISPC_RSYNC_PROGRESS);
				if($f == '/etc/postfix/') $ssh->mv('/tmp/dynamicmaps.cf.migsave', '/etc/postfix/dynamicmaps.cf');
                if(!empty($immutables)) {
                    foreach (explode("\n", $immutables) as $d) {
                        $ssh->exec("chattr +i " . $d);
                    }
                    $immutables = null;
                }
			} elseif(is_file($f)) {
				$this->swriteln('Copy File: '.$f);
				$this->log('Copy File: '.$f);
				$ssh->putFile($f, $f);
			} else {
				// skip as file does not exist
				$this->log('Skip Path: '.$f);
			}
		}
		
		if(!empty($this->web1) && is_array($this->web1)){
			$web1 = $app->db->queryOneRecord("SELECT * FROM web_domain WHERE domain_id = 1");
			if(!empty($web1) && is_array($web1)){
				$server = $app->db->queryOneRecord("SELECT * FROM server WHERE server_id = ?", $this->server_id);
				if($web1['domain'] == $server['server_name'] && $web1['domain'] != $this->web1['domain']){
					$ssh->exec('rm -f '.escapeshellarg(rtrim($server_config['web']['website_basedir'], '/').'/'.$web1['domain']));
				}
			}
		}
	}
	
	// Function to replace IP addresses in ISPConfig
	public function set_replace_ips() {
		global $app, $server_config;

        $app->uses("dbmaster");

		// get all IP addresses
		$ips = array();
		if($server_config['server']['ip_address'] != '127.0.0.1') $ips[] = $server_config['server']['ip_address'];
		
		$server_ip_addresses = $app->db->queryAllRecords("SELECT ip_address FROM server_ip WHERE server_id = ?", intval($this->server_id));
		if(is_array($server_ip_addresses)) {
			foreach($server_ip_addresses as $ip) {
				if(empty($ip['ip_address'])) continue;
				if($ip['ip_address'] == '127.0.0.1') continue;
				if($this->server_role == 'slave' && $this->mirror_server_id > 0){ // dies ist ein Mirror-Server; die IPs des Masters sollen nicht ersetzt werden
					$master_ip = $app->dbmaster->queryOneRecord("SELECT * FROM `server_ip` WHERE `server_id` = ? AND `ip_address` = ?", $this->mirror_server_id, $ip['ip_address']);
					if(!empty($master_ip) && is_array($master_ip)) continue;
				}
				$ips[] = $ip['ip_address'];
			}
		}
		
		// Load IP replace file
		$replace_ip_list = array();
		$replace_ip_file = realpath(BASE_PATH.'/replace-ip-address.txt');
		if(AUTORUN === true) {
			if(file_exists($replace_ip_file) && fileowner($replace_ip_file) === 0) {
				$this->log('Loading IP replacements from file: '.$replace_ip_file);
				$tmps = file($replace_ip_file,FILE_IGNORE_NEW_LINES);
				if(is_array($tmps)) {
					foreach($tmps as $tmp) {
						$parts = explode(' ',$tmp);
						$p0 = trim($parts[0]);
						$p1 = trim($parts[1]);
						if($p0 != '' && $p1 != '') $replace_ip_list[$p0] = $p1;
					}
				}
			}
		}
		
		// Replace IP addresses
		$ips = array_unique($ips);
		foreach ($ips as $ip) {
			// get new IP
			if($ip != '' && $ip != '127.0.0.1') {
				if(AUTORUN === true) {
					if(isset($replace_ip_list[$ip])) {
						$new_ip = $replace_ip_list[$ip];
					} else {
						$new_ip = '';
					}
				} else {
					$this->swriteln($this->lng("Press [Enter] to replace 'OLD_IP' with ['NEW_IP'] or submit a different new IP."));
					$new_ip = $this->free_query('Replace '.$ip.' with',
                        $this->target_server_ip);
				}
				
				$ok = true;
			
				// Check IP address
				if($new_ip != '') {
					if(function_exists('filter_var')) {
						if(!filter_var($new_ip, FILTER_VALIDATE_IP)) {
							$this->swriteln($this->lng('Skipped. Input not a valid IP'));
							$ok = false;
						}
					}
				} else {
					$this->swriteln($this->lng('New IP empty, skipped...'));
					$ok = false;
				}

                if($ok) {
                    $this->writeInputLog("replace_ip_" . preg_replace("/\W/", "_", $ip), $new_ip);
                }
			}
		}
	}

    public function replace_ip() {
        global $ssh, $app, $conf, $server_config;

        $app->uses("dbmaster");

        // Queries to replace IP addresses
        $replace_ip_sql = array();
        $replace_ip_sql_master = array();
        $replace_ip_sql[] = 'UPDATE `web_domain` SET `ip_address` = "{NEWIP}", `generated_vhost` = REPLACE(`generated_vhost`, "{OLDIP}", "{NEWIP}") WHERE `ip_address` = "{OLDIP}" AND `server_id` = {SERVERID}';
        $replace_ip_sql[] = 'UPDATE `web_domain` SET `ipv6_address` = "{NEWIP}", `generated_vhost` = REPLACE(`generated_vhost`, "{OLDIP}", "{NEWIP}") WHERE `ipv6_address` = "{OLDIP}" AND `server_id` = {SERVERID}';
        $replace_ip_sql[] = 'UPDATE `dns_rr` set `data` = replace(`data`, "{OLDIP}", "{NEWIP}") WHERE `data` like "%{OLDIP}% AND `server_id` = {SERVERID}"';
        $replace_ip_sql[] = 'UPDATE `server_ip` SET `ip_address` = "{NEWIP}" WHERE `ip_address` = "{OLDIP}" AND `server_id` = {SERVERID}';

        //$replace_ip_sql[] = 'UPDATE `server` SET `config` = replace(`config`, "{OLDIP}", "{NEWIP}") WHERE `server_id` = {SERVERID}'; // auskommentiert, weil die server-Tabelle nicht übernommen wird
        if($this->server_role == 'slave'){
            $replace_ip_sql_master[] = 'UPDATE `server_ip` SET `ip_address` = "{NEWIP}" WHERE `ip_address` = "{OLDIP}" AND `server_id` = {SERVERID}';
        }

        $ips = $this->getReplaceIps();
        foreach ($ips as $ip => $new_ip) {
            if($ip != '' && $ip != '127.0.0.1') {
                $ok = true;

                // Check IP address
                if($new_ip != '') {
                    if(function_exists('filter_var')) {
                        if(!filter_var($new_ip, FILTER_VALIDATE_IP)) {
                            $this->swriteln($this->lng('Skipped. Input not a valid IP'));
                            $ok = false;
                        }
                    }
                } else {
                    $this->swriteln($this->lng('New IP empty, skipped...'));
                    $ok = false;
                }

                // Run replace commands on target
                if($ok === true) {

                    // Replace placeholders and build combined sql file.
                    $sql_out = '';
                    foreach($replace_ip_sql as $sql) {
                        $sql = str_replace('{NEWIP}',$new_ip,$sql);
                        $sql = str_replace('{OLDIP}',$ip,$sql);
                        $sql = str_replace('{SERVERID}',$this->server_id,$sql);
                        $sql_out .= $sql.";\n";
                    }

                    // copy sql query file to remote server
                    $mysql_dump_file_remote = sys_get_temp_dir().'/replace-ip.sql';
                    $ssh->putFileContents($mysql_dump_file_remote,$sql_out);

                    // Import the dump file
                    $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' '.escapeshellarg($conf['db_database']).' < '.$mysql_dump_file_remote);

                    // cleanup
                    //$ssh->exec('rm -f '.escapeshellarg($mysql_dump_file_remote));

                    // replace IP in nginx configuration
                    //$ssh->exec("perl -p -i -e 's/".$ip."/".$new_ip."/g' /etc/nginx/sites-available/*");

                    if(!empty($replace_ip_sql_master)){
                        foreach($replace_ip_sql_master as $sql_master) {
                            $sql_master = str_replace('{NEWIP}',$new_ip,$sql_master);
                            $sql_master = str_replace('{OLDIP}',$ip,$sql_master);
                            $sql_master = str_replace('{SERVERID}',$this->server_id,$sql_master);
                            $app->dbmaster->query($sql_master);
                        }
                    }

                    $this->swriteln($this->lng('Replaced IP '.$ip.' with '.$new_ip));
                }
            }
        }
    }

    private function getReplaceIps() {
        $ips = [];
        $section = $this->getInputLogSection();

        foreach ($section as $k => $v) {
            preg_match("/^replace_ip_[0-9A-Fa-f\.\:]+/", $k, $matches);
            if(!empty($matches) && $v != "n") {
                if(strpos($v, ":") != false) {
                    $ips[str_replace("_", ":", str_replace("replace_ip_", "", $k))] = $v;
                } else {
                    $ips[str_replace("_", ".", str_replace("replace_ip_", "", $k))] = $v;
                }
            }
        }
        return $ips;
    }
	
	// Function to restart services on the remote system
	public function restart_services($services = 'all') {
		global $ssh, $app, $conf;
		// $services: all | php,postfix,dovecot,amavisd,rspamd,named,pureftpd
		if($services == 'all'){
			$restart_services = array('php', 'postfix', 'dovecot', 'amavisd', 'rspamd', 'named', 'pureftpd');
		} else {
			$restart_services = preg_split('~\s*,\s*~', $services);
		}
		if(empty($restart_services)) return false;
		
		// Debian and Ubuntu
		if(file_exists('/etc/debian_version')) {
			if($app->system->is_installed('apache2') && in_array('php', $restart_services)) {
				$ssh->exec('service apache2 restart');
				$ssh->exec('service service php5-fpm restart');
				$ssh->exec('service service php7.0-fpm restart');
				$ssh->exec('service service php7.3-fpm restart');
				$ssh->exec('service service php7.4-fpm restart');
			}
			if($app->system->is_installed('nginx') && in_array('php', $restart_services)) {
				$ssh->exec('service nginx restart 2>/dev/null');
				$ssh->exec('service service php5-fpm restart');
				$ssh->exec('service service php7.0-fpm restart');
				$ssh->exec('service service php7.3-fpm restart');
				$ssh->exec('service service php7.4-fpm restart');
			}
			if($app->system->is_installed('postfix') && in_array('postfix', $restart_services)) {
				$ssh->exec('service postfix restart');
			}
			if($app->system->is_installed('dovecot') && in_array('dovecot', $restart_services)) {
				$ssh->exec('service dovecot restart');
			}
			if($app->system->is_installed('amavisd-new') && in_array('amavisd', $restart_services)) {
				$ssh->exec('service amavis restart');
			}
			if($app->system->is_installed('rspamd') && in_array('rspamd', $restart_services)) {
				$ssh->exec('service rspamd restart');
			}
			if($app->system->is_installed('named') && in_array('named', $restart_services)) {
				$ssh->exec('service bind9 restart');
			}
			if($app->system->is_installed('pure-ftpd-mysql') && in_array('pureftpd', $restart_services)) {
				$ssh->exec('service pure-ftpd-mysql restart');
			}
		}		
	}
	
	public function adjust_target_config(){
		global $ssh, $app, $conf;
		
		$config_files = array('/usr/local/ispconfig/server/lib/config.inc.php', '/usr/local/ispconfig/interface/lib/config.inc.php');
		
		// adjust server_id
		if($this->server_id != 1){
			foreach($config_files as $config_file){
				// get remote config file
				$tmpfile = tempnam(sys_get_temp_dir(), "config-");
				$ssh->getFile($config_file, $tmpfile);
				
				// extract bind mount lines from local fstab file
				$target_conf = file($tmpfile);
				if(is_array($target_conf)) {
					for($i=0;$i<count($target_conf);$i++) {
						// $conf['server_id'] = '1';
						if(preg_match('@^\$conf\[[\'"]server_id[\'"]\]\s*=\s*[\'"]?\d+[\'"]?;@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'server_id\'] = \''.$this->server_id.'\';'."\n";
						}
					}
				}
				file_put_contents($tmpfile, implode($target_conf));
				
				// Copy fstab file back after doing a backup
				$ssh->cp($config_file, $config_file.'~');
				$ssh->putFile($tmpfile, $config_file);
				
				// cleanup
				unlink($tmpfile);
			}
		}
		
		if($this->server_role == 'slave'){
			// adjust master database credentials
			foreach($config_files as $config_file){
				// get remote config file
				$tmpfile = tempnam(sys_get_temp_dir(), "config-");
				$ssh->getFile($config_file, $tmpfile);
				
				// extract bind mount lines from local fstab file
				$target_conf = file($tmpfile);
				if(is_array($target_conf)) {
					for($i=0;$i<count($target_conf);$i++) {
						// $conf['dbmaster_host']
						if(preg_match('@^\$conf\[[\'"]dbmaster_host[\'"]\]\s*=\s*[\'"].*[\'"];@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'dbmaster_host\']                  = \''.$conf['dbmaster_host'].'\';'."\n";
						}
						// $conf['dbmaster_port']
						if(preg_match('@^\$conf\[[\'"]dbmaster_port[\'"]\]\s*=\s*[\'"].*[\'"];@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'dbmaster_port\']                  = \''.(!empty($conf['dbmaster_port'])? $conf['dbmaster_port'] : '3306').'\';'."\n";
						}
						// $conf['dbmaster_database']
						if(preg_match('@^\$conf\[[\'"]dbmaster_database[\'"]\]\s*=\s*[\'"].*[\'"];@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'dbmaster_database\']              = \''.$conf['dbmaster_database'].'\';'."\n";
						}
						// $conf['dbmaster_user']
						if(preg_match('@^\$conf\[[\'"]dbmaster_user[\'"]\]\s*=\s*[\'"].*[\'"];@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'dbmaster_user\']                  = \''.$conf['dbmaster_user'].'\';'."\n";
						}
						// $conf['dbmaster_password']
						if(preg_match('@^\$conf\[[\'"]dbmaster_password[\'"]\]\s*=\s*[\'"].*[\'"];@',$target_conf[$i])){
							$target_conf[$i] = '$conf[\'dbmaster_password\']              = \''.$conf['dbmaster_password'].'\';'."\n";
						}
					}
				}
				file_put_contents($tmpfile, implode($target_conf));
				
				// Copy fstab file back after doing a backup
				$ssh->cp($config_file, $config_file.'~');
				$ssh->putFile($tmpfile, $config_file);
				
				// cleanup
				unlink($tmpfile);
			}
			
			// adjust server_id in Dovecot, Postfix, and PureFTPd configuration files
			// preg_replace('~(and\s+(w+\.)?server_id)\s*=\s*(\')?(\d+)(\')?($|\s+)~i','$1 = ${3}'.$this->server_id.'${5}${6}',$sourcestring);
			$config_files = array(
									array('file' => '/etc/dovecot/dovecot-sql.conf', 'perms' => 0600),
									array('file' => '/etc/postfix/mysql-virtual_domains.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_email2email.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_forwardings.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_gids.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_mailboxes.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_recipient.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_relaydomains.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_relayrecipientmaps.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_sender.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_transports.cf', 'perms' => 0640),
									array('file' => '/etc/postfix/mysql-virtual_uids.cf', 'perms' => 0640),
									array('file' => '/etc/pure-ftpd/db/mysql.conf', 'perms' => 0600),
									);
			foreach($config_files as $config_file){
				// get remote config file
				$tmpfile = tempnam(sys_get_temp_dir(), "config-");
				if(empty($config_file['perms'])){
					$config_file['perms'] = 0644;
				}
				chmod($tmpfile, $config_file['perms']);
				$ssh->getFile($config_file['file'], $tmpfile);
				
				// extract bind mount lines from local fstab file
				$target_conf = file($tmpfile);
				if(is_array($target_conf)) {
					for($i=0;$i<count($target_conf);$i++) {
						$target_conf[$i] = preg_replace('~((?:and|or)\s+(\w+\.)?server_id)\s*=\s*(\')?(\d+)(\')?($|\s+)~i','$1 = ${3}'.$this->server_id.'${5}${6}', $target_conf[$i]);
					}
				}
				file_put_contents($tmpfile, implode($target_conf));
				
				// Copy file back after doing a backup
				$ssh->cp($config_file['file'], $config_file['file'].'~');
				$ssh->putFile($tmpfile, $config_file['file']);
				
				// cleanup
				unlink($tmpfile);
			}
			// Restart services on remote system
			$this->swriteln($this->lng('Restarting Dovecot, Postfix, and PureFTPd on target system...'));
			$this->restart_services('dovecot,postfix,pureftpd');
		}
	}
	
	public function master_change_mysql_access(){
		global $app, $conf, $master_ssh, $ssh;
		
		if(empty($this->master_server)) return false;
		// falls Slave, ändere MySQL-Zugriffsrechte auf dem Master
		// Servername
		$master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($conf['dbmaster_database']) . ' -e "UPDATE server SET server_name = \''.$this->target_server_hostname.'\' WHERE server_id = '.$this->server_id.'"');
		// IP-Adresse
		$master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($conf['dbmaster_database']) . ' -e "UPDATE server_ip SET ip_address = \''.$this->target_server_ip.'\' WHERE ip_address = \''.$this->local_server_ip.'\' AND server_id = '.$this->server_id.'"');
		if($this->server_role == 'slave'){
			$ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($conf['db_database']) . ' -e "UPDATE server_ip SET ip_address = \''.$this->target_server_ip.'\' WHERE ip_address = \''.$this->local_server_ip.'\' AND server_id = '.$this->server_id.'"');
		}
		$master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($conf['dbmaster_database']) . ' -e "UPDATE web_domain SET ip_address = \''.$this->target_server_ip.'\' WHERE ip_address = \''.$this->local_server_ip.'\' AND server_id = '.$this->server_id.'"');
		// PHP-Versionen
//		if(!empty($this->target_php_versions) && is_array($this->target_php_versions)){
//			$php_versions_to_keep = array();
//			foreach($this->target_php_versions as $php_version){
//				if(!in_array($php_version['php_fpm_init_script'], $php_versions_to_keep)) $php_versions_to_keep[] = $php_version['php_fpm_init_script'];
//			}
//			if(!empty($php_versions_to_keep)){
//				$php_fpm_init_script_str = '';
//				foreach($php_versions_to_keep as $php_version_to_keep){
//					$php_fpm_init_script_str .= "'".$php_version_to_keep."',";
//				}
//				$php_fpm_init_script_str = rtrim($php_fpm_init_script_str, ',');
//				$cmd = 'mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D '.escapeshellarg($conf['dbmaster_database']) . ' -e "DELETE FROM server_php WHERE server_id = '.$this->server_id.' AND php_fpm_init_script NOT IN('.$php_fpm_init_script_str.')"';
//				$master_ssh->exec($cmd);
//			}
//		}
//
//        // columns_priv
//        $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE columns_priv SET Host = \'' . $this->target_server_hostname . '\' WHERE Host = \'' . $this->local_server_hostname . '\'"');
//        $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE columns_priv SET Host = \'' . $this->target_server_ip . '\' WHERE Host = \'' . $this->local_server_ip . '\'"');
//
//        // tables_priv
//        $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE tables_priv SET Host = \'' . $this->target_server_hostname . '\' WHERE Host = \'' . $this->local_server_hostname . '\'"');
//        $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE tables_priv SET Host = \'' . $this->target_server_ip . '\' WHERE Host = \'' . $this->local_server_ip . '\'"');
//
//        $check_mysql_user_for_table = $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "SHOW CREATE TABLE user;" | grep -i "create table"');
//        if(empty($check_mysql_user_for_table)) {
//            $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE global_priv SET host = \'' . $this->target_server_hostname . '\' WHERE host = \'' . $this->local_server_hostname . '\'"');
//            $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE global_priv SET host = \'' . $this->target_server_ip . '\' WHERE host = \'' . $this->local_server_ip . '\'"');
//        } else {
//            // user
//            $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE user SET Host = \'' . $this->target_server_hostname . '\' WHERE Host = \'' . $this->local_server_hostname . '\'"');
//            $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "UPDATE user SET Host = \'' . $this->target_server_ip . '\' WHERE Host = \'' . $this->local_server_ip . '\'"');
//
//        }
//        $master_ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "FLUSH PRIVILEGES;"');
	}
	
	public function new_master_adjust_slaves(){
		global $app, $conf, $ssh;
		
		if($this->server_role != 'master') return false;
		
		$db_name = 'mysql';
		$tables = array('columns_priv', 'tables_priv');
        $check_mysql_user_for_table = $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "SHOW CREATE TABLE user;" | grep -i "create table"');
        exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "SHOW CREATE TABLE user;" | grep -i "create table"', $check_source_mysql_user_for_table);

        if(empty($check_mysql_user_for_table) && empty($check_source_mysql_user_for_table)) {
            $tables[] = "global_priv";
        } else {
            $tables[] = "user";
        }

		foreach($tables as $table){
			$mysql_dump_file = tempnam(sys_get_temp_dir(), "mysql-");
			$mysql_dump_file2 = tempnam(sys_get_temp_dir(), "mysql-");
			if($table == 'user'){
				$where = '--where="User LIKE \'ispcsrv%\'"';
			} elseif ($table == 'global_priv') {
                $where = '--where="user LIKE \'ispcsrv%\'"';
            } else {
				$where = '--where="Db = \''.$conf['db_database'].'\'"';
			}
			if(!empty($where)) $where .= ' --skip-add-drop-table --no-create-info ';
			
			//$cmd = 'mysqldump --defaults-extra-file='.$this->mysql_defaults_extra_file.' -cCeQ -f --hex-blob --quote-names --quick --routines --events --triggers '.$where.escapeshellarg($db_name) .' '. escapeshellarg($table) . ' >> '.escapeshellarg($mysql_dump_file2);
			$cmd = 'mysqldump --defaults-extra-file='.$this->mysql_defaults_extra_file.' -cCeQ -f --hex-blob --quote-names --quick --max_allowed_packet=512M '.$where.escapeshellarg($db_name) .' '. escapeshellarg($table) . ' >> '.escapeshellarg($mysql_dump_file2);
			//echo $cmd."\n";
			exec($cmd);
			
			exec('cat '.escapeshellarg($mysql_dump_file2).' | gzip -6 > '.escapeshellarg($mysql_dump_file));
			unlink($mysql_dump_file2);
			chmod($mysql_dump_file,0600);
			
			$this->log('Executed local command '.$cmd,0);
			
			// Copy the dump file to the target
			$ssh->putFile($mysql_dump_file,$mysql_dump_file);
			
			// Import the dump file
			$ssh->exec('zcat ' . escapeshellarg($mysql_dump_file) . ' | mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -f --max_allowed_packet=512M -D '.escapeshellarg($db_name));
			
			// cleanup
			$ssh->exec('rm -f '.escapeshellarg($mysql_dump_file));
			unlink($mysql_dump_file);
		}
		
		$ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D mysql -e "FLUSH PRIVILEGES;"');
		
		// auf Slave-Servern den neuen Master-Server in beiden config.inc.php-Dateien anpassen
		$server = $app->db->queryOneRecord("SELECT * FROM server WHERE server_id = ?", $this->server_id);
		if(!empty($server) && is_array($server) && $server['server_name'] != $this->target_server_hostname){
			$sleep = 90;
			$this->swriteln($this->lng('Changing master server name from '.$server['server_name'].' to '.$this->target_server_hostname.' and updating master server hostname in both config.inc.php files on slave servers. Waiting '.$sleep.' seconds for operation to complete...'));
			$server['server_name'] = $this->target_server_hostname;
			$app->db->datalogUpdate('server', $server, 'server_id', $this->server_id); // this updates server name on mirror servers automatically
			
			$other_servers = $app->db->queryAllRecords("SELECT * FROM server WHERE server_id != ? AND mirror_server_id = 0", $this->server_id);
			if(!empty($other_servers) && is_array($other_servers)){
				foreach($other_servers as $other_server){
					$app->db->datalogUpdate('server', $other_server, 'server_id', $other_server['server_id'], true);
				}
			}
			sleep($sleep);
		}
	}
	
	public function target_disable_ispconfig_interface(){
		global $ssh, $app, $conf;
		
		if(empty($this->master_server)) return false;
		// falls Slave, deaktiviere ISPConfig-Vhost
		if(!$app->system->is_installed('nginx')) return false;
		
		$ssh->exec('rm -f /etc/nginx/sites-enabled/000-ispconfig.vhost 2>/dev/null');
		$ssh->exec('rm -f /etc/nginx/sites-available/ispconfig.vhost 2>/dev/null');
		$ssh->exec('service nginx restart 2>/dev/null');
		return true;
	}
	
	public function is_docker_available(){
		global $ssh;
		
		exec('which docker', $output, $retval);
		if(!empty($retval)){
			$this->log(__FUNCTION__ . ': docker is not installed on source server.');
			return false;
		}
		unset($output);
		
		exec('which docker-compose', $output, $retval);
		if(!empty($retval)){
			$this->log(__FUNCTION__ . ': docker-compose is not installed on source server.');
			return false;
		}
		unset($output);
		
		$output = $ssh->exec('which docker');
		if($output === false){
			$this->log(__FUNCTION__ . ': docker is not installed on target server.');
			return false;
		}
		unset($output);
		
		$output = $ssh->exec('which docker-compose');
		if($output === false){
			$this->log(__FUNCTION__ . ': docker-compose is not installed on target server.');
			return false;
		}
		unset($output);
		
		return true;
	}
	
	public function start_docker_containers(){
		global $ssh, $app, $conf;
		
		if(!$this->is_docker_available()) return false;
		
		$sql = "SELECT * FROM web_domain WHERE server_id = ? AND type = 'vhost' AND active = 'y'";
		$records = $app->db->queryAllRecords($sql, $conf['server_id']);
		if(empty($records) || !is_array($records)) return false;
		
		foreach($records as $rec) {
			$web_path = $rec['document_root'];
			if(!is_dir($web_path . '/misc')) continue; // nothing to back up
			
			$web_id = $rec['domain_id'];
			
			$docker_dir = $web_path.'/misc/web*/docker/';
			foreach (glob($docker_dir.'*/docker-compose.yml') as $filename){
				$filename = realpath($filename);
				if($filename === false) continue;
				
				$cmd = 'docker-compose -f '.escapeshellarg($filename).' up -d 2>/dev/null';
				$ssh->exec($cmd);
			}
		}
	}
	
	public function migrate_docker_databases(){
		global $ssh, $app, $conf;
		
		if(!$this->is_docker_available()) return false;
		
		// find Docker databases
		$docker_dbs = $app->db->queryAllRecords("SELECT * FROM `web_database` WHERE `database_version` != '' AND `server_id` = ?", (int)$conf['server_id']);
		if(empty($docker_dbs) || !is_array($docker_dbs)){
			$this->log(__FUNCTION__ . ': no Docker databases found to be migrated to target server.');
			return false;
		}
		
		// make sure /opt/db/ exists on target server
		$docker_base_path = '/opt/db/';
		if(!$ssh->isDir($docker_base_path)){
			$this->log(__FUNCTION__ . ': creating Docker base directory '.$docker_base_path.' on target server.');
			$ssh->mkdir($docker_base_path, '0750');
			$cmd = 'chown root:root '.escapeshellarg($docker_base_path);
			$ssh->exec($cmd);
		}
		
		$count = 0;
		foreach($docker_dbs as $docker_db){
			$docker_path = $docker_base_path.'db'.$docker_db['database_id'].'/';
			$file = $docker_path.'docker-compose.yml';
			if(!file_exists($file)){
				$this->log(__FUNCTION__ . ': docker-compose file '.$file.' does not exist on source server, skipping migration of Docker database ID ['.$docker_db['database_id'].'].');
				continue;
			}
			// stopping database container on source server
			$cmd = 'cd '.escapeshellarg($docker_path).' && docker-compose -f '.escapeshellarg($file).' down 2>/dev/null';
			exec($cmd, $output, $retval);
			if(!empty($retval)){
				$this->log(__FUNCTION__ . ': could not stop Docker database ID ['.$docker_db['database_id'].'] on source server.'.(!empty($output)? 'Error: '.print_r($output, true) : ''));
				unset($output);
				continue;
			}
			unset($output);
			
			// copying Docker directory to target server
			$this->swriteln('Copy Directory: '.$docker_path);
			$this->log(__FUNCTION__ . ': copy Directory: '.$docker_path);
			$docker_path = rtrim($docker_path, '/').'/';
			$ssh->copyFiles($docker_path, $docker_path, 'rsync', true, ISPC_RSYNC_PROGRESS);
			
			// starting database container on target server
			$cmd = 'cd '.escapeshellarg($docker_path).' && docker-compose -f '.escapeshellarg($file).' up -d 2>/dev/null';
			$output = $ssh->exec($cmd);
			if($output === false){
				$this->log(__FUNCTION__ . ': could not start Docker database ID ['.$docker_db['database_id'].'] on target server.');
			} else {
				$this->log(__FUNCTION__ . ': successfully migrated and started Docker database ID ['.$docker_db['database_id'].'] to target server.');
			}
			$this->log(__FUNCTION__ . ': leaving Docker database ID ['.$docker_db['database_id'].'] stopped on source server. If needed you have to start it manually on the source system. BEWARE: this will lead to data not being in sync on source and target server!!!');
			$count++;
		}
		$this->log(__FUNCTION__ . ': migrated '.$count.' Docker database'.($count == 1? '' : 's').'.');
		return true;
	}

    public function setDockerRootPw() {
        global $ssh;

        $base_docker_dir = "/var/lib/docker/containers";
        $old_pw = file_get_contents("/usr/local/ispconfig/server/lib/mysql_root_pw.env");
        $old_pw = explode("=", $old_pw)[1];
        $new_pw = $ssh->getFileContents("/usr/local/ispconfig/server/lib/mysql_root_pw.env");
        $new_pw = explode("=", $new_pw)[1];
        $tmp = $ssh->exec("ls $base_docker_dir");
        $docker_folders = [];
        foreach (explode("\n", $tmp) as $row) {
            foreach (explode("\t", $row) as $folder) {
                $docker_folders[] = $folder;
            }
        }
        $docker_ports = [];
        foreach ($docker_folders as $f) {
            if($f != "." && $f != ".." && !empty($f)) {
                $conf = json_decode($ssh->getFileContents("$base_docker_dir/$f/config.v2.json"), true);
                if(array_key_exists("3306/tcp", $conf["NetworkSettings"]["Ports"])) {
                    if (!in_array($conf["NetworkSettings"]["Networks"]["db5_default"]["IPAddress"], $docker_ports)) {
                        $docker_ports[] = $conf["NetworkSettings"]["Ports"]["3306/tcp"][0]["HostPort"];
                    }
                }
            }
        }
        foreach ($docker_ports as $docker_port) {
            $check_old_pw = $ssh->exec("mysql -h 127.0.0.1 -u root -p$old_pw --port $docker_port -e \"SHOW DATABASES;\"");
            if(strpos($check_old_pw, "Access denied for user") != false) {
                $sql = "mysql -h 127.0.0.1 -u root -p$old_pw --port $docker_port -e \"ALTER USER 'root'@'%' IDENTIFIED BY '$new_pw';\"";
                $ssh->exec($sql);
            }
        }
    }
    
    public function copy_crons_to_tmp() {
        global $ssh;
        
        if(!$ssh->isDir("/tmp")) {
            $ssh->exec("mkdir /tmp");
            $ssh->exec("chmod 0777 /tmp");
        }
        if(!$ssh->isDir("/tmp/ispconfig_crons")) {
            $ssh->exec("mkdir /tmp/ispconfig_crons");
            $ssh->exec("chmod 0777 /tmp/ispconfig_crons");
        }
    
        $crons = glob("/etc/cron.d/ispc_*");
        if (!empty($crons) && is_array($crons)) {
            foreach ($crons as $cron) {
                $ssh->putFile($cron, "/tmp/ispconfig_crons/" . str_replace("/etc/cron.d/", "", $cron));
            }
        }
    }

    /**
     * Disables all User crons on source and if is slave on master server by setting them to active = 'n'
     *
     * @return bool
     */
    public function stopUserCrons() {
        if($this->getInputLogEntry("stop_user_crons") == "y") {
            $this->swriteln("Stop user crons...");
            exec("/opt/th-ispcopy/lib/rp-maint/rp-maint.sh --clearcrons", $ouptut);
            $this->swriteln(implode("\n", $ouptut));
            return false;
        }
        return false;
    }

    public function toggleUserCronsOnTarget() {
        global $ssh;

        if($this->getInputLogEntry("mode") == "first") {
            $source = "/etc/cron.d/ispc_*";
            $target = "/root/ispcopy/user-crons-bak/";
        } else if($this->getInputLogEntry("mode") == "second") {
            $source = "/root/ispcopy/user-crons-bak/ispc_*";
            $target = "/etc/cron.d/";
        }
        if(!empty($source) && !empty($target)) {
            $check = $ssh->exec("ls -l $source 2>/dev/null");
            if(!empty($check)) {
                $ssh->exec("mkdir -p $target");
                $ssh->exec("mv $source $target");
            }
        }
        return false;
    }

    /**
     * Stops all system cron jobs by comment every row in crontab
     *
     * @return void
     */
    public function stopSysCrons() {
        if($this->getInputLogEntry("stop_sys_crons") == "y") {
            exec("crontab -l > sys_crons.tmp");
            $crons = explode("\n", file_get_contents("sys_crons.tmp"));
            for ($i = 0; $i < count($crons); $i++) {
                if (substr(trim($crons[$i]), 0, 1) != "#") {
                    $crons[$i] = "#" . $crons[$i];
                }
            }
            file_put_contents("sys_crons.tmp", implode("\n", $crons));
            exec("crontab sys_crons.tmp");
        }
    }

    /**
     * Configures a proxy on the source server to the target server
     *
     * @return bool
     */
    public function setProxy() {
        if($this->getInputLogEntry("set_up_proxy") == "y") {
            $this->swriteln("Set up proxy...");

            exec("/opt/th-ispcopy/lib/rp-maint/rp-maint.sh --ispcopy_ini", $ouptut);
            $this->swriteln(implode("\n", $ouptut));
            return false;
        }
        return false;
    }

    public function doLostSlavePostWork() {
        $this->swriteln("Do post work for slave server.");
        $this->swriteln("HINT: Outputs will be suppressed, because initializing \$app causes errors on master, where the source slave has no longer access to.");
        exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/enable_reverse_proxy_on_lost_slave.php --ip={$this->target_server_ip} > /dev/null 2>&1");
    }

    /**
     * Activates mail routing on the source server to the target server
     *
     * @return bool
     */
    public function setMailRouting() {
        if($this->getInputLogEntry("activate_routing") == "y") {
            $this->swriteln("Activate mail routing...");
            exec("/opt/th-ispcopy/lib/rp-maint/rp-maint.sh --routing", $ouptut);
            $this->swriteln(implode("\n", $ouptut));
            return false;
        }
        return false;
    }

    /**
     * Activates sni on the target server
     *
     * @return bool
     */
    public function activateSNI() {
        global $ssh;
        if($this->getInputLogEntry("activate_sni") == "y") {
            $this->swriteln("Activate SNI...");
            $ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --update-server-config=server.use_sni:y");
            $this->runServerSh("target");
            return true;
        }
        return false;
    }

    /**
     * Creates servername with old hostname if sni is activated
     *
     * @return bool
     */
    public function createServerName() {
        global $ssh, $master_ssh, $conf;
        $tmp_ssh = $ssh;
        if(!empty($this->master_server)) {
            $tmp_ssh = $master_ssh;
        }
        if($this->getInputLogEntry("activate_sni") == "y") {
            $db_name = $conf['db_database'];
            if(!empty($this->master_server)) {
                $db_name = $conf['dbmaster_database'];
            }

            $check = $tmp_ssh->exec("mysql --defaults-extra-file=/etc/mysql/debian.cnf $db_name -sNe \"SELECT * FROM server_name WHERE server_name = '" . $this->local_server_hostname . "'\"");
            if(empty($check)) {
                $this->swriteln("Create server name...");
                $tmp_ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --create-server-name=serverId:{$this->server_id},mailActive:1,serverName:{$this->local_server_hostname}");
                return true;
            }
        }
        return false;
    }

    /**
     * Using netcat to check if the target server is on the Telekom blacklist
     *
     * @return bool
     */
    public function isOnTelekomBlacklist() {
        global $ssh;
    
        $result = $ssh->exec("timeout -k 1 1 netcat mx01.t-online.de 25");
        return trim(explode(" ", $result[0])[0]) != "220";
    }

    /**
     * Activates maintenance mode for all user sites on the source server and saves where the maintenance mode has been allready activated
     *
     * @return bool
     */
    public function activate_user_maintenance() {
        if($this->getInputLogEntry("user_maintenance_on") == "y") {
            $this->swriteln("Activate maintenance mode of user sites...");
            exec("/opt/th-ispcopy/lib/rp-maint/rp-maint.sh --maint", $output);
            $this->swriteln(implode("\n", $output));
            $this->writeInputLog("user_maintenance_off", "n");
            return false;
        } else {
            $this->writeInputLog("user_maintenance_off", "n");
        }
        return false;
    }

    /**
     * Disables maintenance mode for all sites, where the meintanenace mode has been enabled before deactivating by activate_user_maintenance
     *
     * @return bool
     */
    public function disable_user_maintenance() {
        global $ssh, $master_ssh;

        if($this->getInputLogEntry("user_maintenance_off") == "y") {
            $this->swriteln("Disable maintenance of user sites on new (target) server...");
            $maintenance_updated = false;
            $maintenance_master_updated = false;
            $vhosts = $ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-vhosts=server_id:{$this->server_id} --raw");
            foreach (explode("\n", $vhosts) as $vhost) {
                $id = explode(",", $vhost)[0];
                if ($this->getInputLogEntry($id, "maintenance") == "n") {
                    $maintenance_updated = true;
                    $ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --update-vhost=maintenance_mode:0 --where=id:$id");
                }
            }
            if(!empty($this->master_server)) {
                $vhosts_master = $master_ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --show-vhosts=server_id:{$this->server_id} --raw");
                foreach (explode("\n", $vhosts_master) as $vhost_master) {
                    $id = explode(",", $vhost_master)[0];
                    if ($this->getInputLogEntry($id, "maintenance") == "n") {
                        $maintenance_master_updated = true;
                        $master_ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --update-vhost=maintenance_mode:0 --where=id:$id");
                    }
                }
            }
            if ($maintenance_updated) {
                return true;
            }
            if ($maintenance_master_updated) {
                return true;
            }
        }
        return false;
    }

    /**
     * Replaces all IP adresses and domains on target server, where source ip or domain exists in `web_domain` table.
     *
     * @return bool
     */
    public function replaceServernameInAliasdomains() {
        global $ssh, $master_ssh, $conf;

        $this->swriteln("Replace old server names in aliasdomains on the target system");
        $updated = false;
        $domains = explode("\n", $ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." " . $conf['dbmaster_database'] . " -sNe \"SELECT ip_address, domain FROM web_domain WHERE server_id = " . $this->server_id . ";\""));
        $updated = $this->replaceDomains($domains);
        if($this->server_role == "slave") {
            $domains = explode("\n", $master_ssh->exec("mysql --defaults-extra-file=/etc/mysql/debian.cnf " . $conf['dbmaster_database'] . " -sNe \"SELECT ip_address, domain FROM web_domain WHERE server_id = " . $this->server_id . ";\""));
            $updated = $this->replaceDomains($domains, "master");
        }
        if(!$updated) {
            $link_results = $ssh->exec("find /var/www/ /var/www/clients/client*/ -maxdepth 1 -type l -regex '.*/\(.*\.\)?" . str_replace(".", "\\.", $this->local_server_hostname) . "$'");
            if(!empty($link_results)) {
                $updated = "";
                foreach (explode("\n", $link_results) as $result) {
                    $check = $ssh->exec("ls " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result));
                    if(!empty($check) && strpos($check, "No such file or directory") == false) {
                        $updated .= "unlink " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result) . " && ";
                    }
                    $updated .= "rsync -a " . $result . " " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result) . " && unlink " . $result . "\n";
                }
            }
            $dir_results = $ssh->exec("find /var/log/ispconfig/httpd/ -maxdepth 1 -type d -regex '.*/\(.*\.\)?" . str_replace(".", "\\.", $this->local_server_hostname) . "$'");
            if(!empty($dir_results)) {
                if(is_bool($updated)) {
                    $updated = "";
                }
                foreach (explode("\n", $dir_results) as $result) {
                    $source_immutable = $ssh->exec("lsattr -d " . $result . " 2>/dev/null | grep -- '-i-'");
                    $target_immutable = $ssh->exec("lsattr -d " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result) . " 2>/dev/null | grep -- '-i-'");
                    if(!empty($source_immutable)) {
                        $updated .= "chattr -i " . $result . " && ";
                    }
                    if(!empty($target_immutable)) {
                        $updated .= "chattr -i " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result) . " && ";
                    }
                    $updated .= "rsync -a " . $result . "/ " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result) . " && rm -rf " . $result;
                    if(!empty($target_immutable)) {
                        $updated .= " && chattr +i " . str_replace($this->local_server_hostname, $this->target_server_hostname, $result);
                    }
                    $updated .= "\n";
                }
                $updated = trim($updated);
            }
        }
        return $updated;
    }

    private function replaceDomains($domains, $target = "target") {
        global $ssh, $master_ssh;

        $updated = false;
        $tmp_ssh = $ssh;
        if($target == "master") {
            $tmp_ssh = $master_ssh;
        }
        foreach($domains as $row) {
            if(count(explode("\t", $row)) >= 2) {
                $domain = trim(explode("\t", $row)[1]);

                preg_match("~(^|\.){$this->local_server_hostname}$~", $domain, $matches);
                if (!empty($matches)) {
                    $new_subdomain = str_replace($this->local_server_hostname, $this->target_server_hostname, $domain);
                    $this->swriteln("Replace sub-/alias domain {$domain} by {$new_subdomain}.");
                    $tmp_ssh->exec("/usr/local/bin/ispcphp /usr/local/ispconfig/server/scripts/console/ISPCConsole.php --update-vhost=domain:\"{$new_subdomain}\" --where=domain:\"{$domain}\",server_id=\"{$this->server_id}\"");
                    $updated = true;
                }
            }
        }
        return $updated;
    }

    public function resetSlavesSysDatalogOnSecondTransfer($mode) {
        global $conf, $ssh, $master_ssh;

        if(!empty($master_ssh)) {
            $sql_max_datalog = "SELECT MAX(datalog_id) FROM sys_datalog WHERE server_id = {$this->server_id};";
            $sql_last_updated = "SELECT MAX(datalog_id) FROM sys_datalog WHERE server_id = {$this->server_id} AND status <> 'pending';";
            $max_datalog = (int)$master_ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['dbmaster_database']} -sNe \"$sql_max_datalog\"");
            $max_datalog += 1;
            $last_updated = (int)$master_ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['dbmaster_database']} -sNe \"$sql_last_updated\"");
            $sql_set_updated = "UPDATE server SET updated = $last_updated WHERE server_id = {$this->server_id}";
            $sql_remove_sys_datalog_entries = "DELETE FROM sys_datalog WHERE server_id = {$this->server_id} AND datalog_id > $max_datalog";
            $sys_datalog_auto_increment = "ALTER TABLE sys_datalog AUTO_INCREMENT = {$max_datalog};";

            $ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -e \"{$sql_set_updated}\"");
            $ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -e \"{$sql_remove_sys_datalog_entries}\"");
            $ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -e \"{$sys_datalog_auto_increment}\"");
        }
    }

    public function replaceConfig($type = "target") {
        global $ssh, $master_ssh, $conf;

        if($type == "master" && empty($this->master_server)) {
            return false;
        }

        $tmp_ssh = $ssh;
        if($type == "master") {
            $tmp_ssh = $master_ssh;
        }
        $sql_update_ip = "UPDATE server SET config = REPLACE(config, '{$this->local_server_ip}', '{$this->target_server_ip}') WHERE server_id = {$this->server_id}";
        $sql_update_hostname = "UPDATE server SET config = REPLACE(config, '{$this->local_server_hostname}', '{$this->target_server_hostname}') WHERE server_id = {$this->server_id}";
        $tmp_ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -e \"{$sql_update_ip}\"");
        $tmp_ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -e \"{$sql_update_hostname}\"");
    }

    public function copySSL() {
        global $ssh;

        exec("find /usr/local/ispconfig/interface/ssl -regex '.*key\|.*crt'", $files);
        if(!empty($files)) {
            foreach ($files as $file) {
                if(!$ssh->isFile($file)) {
                    $ssh->putFile($file, $file);
                }
            }
        }
        if($ssh->isDir("/root/ispcopy/default-vhost-ssl") && $ssh->isDir("/var/www/clients/client1/web1")) {
            $ssh->exec("mv /root/ispcopy/default-vhost-ssl/* /var/www/clients/client1/web1/ssl/");
        }
    }

    /**
     * Detects the user that logged in via ssh
     *
     * @return mixed|string
     */
    public function getTimmeUser() {
        exec("echo \$SUDO_USER", $user);
        if(empty($user)) {
            return "";
        } else {
            return $user[0];
        }
    }


    public function askQuestions($mode) {
        $ignore_input_file = $mode == "second" && $this->getInputLogEntry("mode") == "first";
        if($mode == "second") {
            $this->writeInputLog("mode", "second");
            if(!empty($this->master_server)) {
                $slave_second_mode = $this->simple_query("Shall the slave be copied as slave, or shall it be a standalone server?",
                    ["slave", "standalone"],
                    "slave",
                    "slave_second",
                    true);
                if($slave_second_mode == "standalone") {
                    $this->master_server = null;
                    $this->server_role = "single";
                }
            }
        }
        $activate_user_maintenance = $this->simple_query("Shall the maintenance mode be activated for user sites?", ["y", "n"], $mode == "first"? "n":"y", "user_maintenance_on", $ignore_input_file);
        $this->simple_query('Shall the user cronjobs be stopped on source (old) server?', array('y','n'), $mode == "first"? "n":"y", 'stop_user_crons', $ignore_input_file);
        $this->simple_query('Shall the system cronjobs be stopped on source (old) server?', array('y','n'), $mode == "first"? "n":"y", 'stop_sys_crons', $ignore_input_file);
//        if($activate_user_maintenance == "y") {
//            $this->simple_query("Shall the maintenance mode on target (new) server be disabled for user sites after copy?", ["y", "n"], $mode == "first"? "n":"y", "user_maintenance_off", $ignore_input_file);
//        }
        $this->simple_query("Should a proxy be set up?", ["y", "n"], $mode == "first"? "n":"y", "set_up_proxy", $ignore_input_file);
        $this->simple_query("Should e-mail routing be activated?", ["y", "n"], $mode == "first"? "n":"y", "activate_routing", $ignore_input_file);
        if($this->getInputLogEntry("activate_sni") == "y") {
            $this->writeInputLog("activate_sni", "n");
        } else {
            $this->simple_query("Activate SNI?", ["y", "n"], "y", "activate_sni", $ignore_input_file);
        }
        if($mode == "second") {
            $this->simple_query("Run resync after ispcopy?", ["y", "n"], "y", "run_resync", $ignore_input_file);
        } else {
            $this->writeInputLog("run_resync", "y");
        }
        $this->set_replace_ips();
    }

    /**
     * Genrates a list with all new hossts and writes them into '/root/ispcopy_new_hosts'
     *
     * @return void
     */
    public function generateHostList() {
        global $ssh;

        $hosts = $this->getHosts();
        $file_content = "";
        foreach ($hosts as $k => $v) {
            $file_content .= $k . " " . implode(" ", $v) . "\n";
        }

        $hosts_file = "{$this->getLogDir()}/ispcopy_new_hosts";

        file_put_contents($hosts_file, str_replace("NULL", "    ", trim($file_content) . "\n"));

        $this->swriteln("Wrote list of all hosts to '$hosts_file'.");
    }

    private function getHosts() {
        global $ssh, $conf;

        $hosts = [];
        $sql_ips = "SELECT DISTINCT `ip_address` FROM `web_domain` WHERE `parent_domain_id` = 0;";
        $ip_data = $ssh->exec("mysql --defaults-extra-file=".$this->mysql_defaults_extra_file." -D {$conf['db_database']} -Nse '$sql_ips'");
        if(!empty($ip_data)) {
            foreach (explode("\n", $ip_data) as $ip) {
                $sql = <<<SQL
SELECT DISTINCT
        domain,
        CONCAT_WS('', case subdomain when 'none' then '' else 'www.' end, domain)
FROM
        web_domain
WHERE
        ip_address = '$ip';
SQL;
                $domain_data = $ssh->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -D ' . $conf['db_database'] . ' -Nse "' . $sql . '"');
                $domains = [];
                foreach (explode("\n", $domain_data) as $domain) {
                    $domains[] = explode("\t", $domain)[0];

                    $domains[] = explode("\t", $domain)[1];
                }
                $hosts[$ip] = array_unique($domains);
            }
        }

        return $hosts;
    }

    public function replaceServerId() {
        global $ssh, $conf;

        if($this->getInputLogEntry("slave_second") == "standalone") {
            if($this->server_id != 1) {
                $this->swriteln("Set server_id to 1...");
                $ssh->exec("mysql --defaults-extra-file=" . $this->mysql_defaults_extra_file . " " . $conf['db_database'] . " -e 'UPDATE server SET server_id = 1;'");
                $tables = $ssh->exec("mysql --defaults-extra-file=" . $this->mysql_defaults_extra_file . " " . $conf['db_database'] . " -sNe 'SHOW TABLES;'");
                foreach (explode("\n", $tables) as $table) {
                    if($table != "server") {
                        if (!empty(trim($ssh->exec("mysql --defaults-extra-file=" . $this->mysql_defaults_extra_file . " " . $conf['db_database'] . " -sNe 'SHOW CREATE TABLE " . $table . ";' | grep '`server_id`'")))) {
                            $ssh->exec("mysql --defaults-extra-file=" . $this->mysql_defaults_extra_file . " " . $conf['db_database'] . " -e 'UPDATE " . $table . " SET server_id = 1;'");
                        }
                    }
                }
                $config_content = $ssh->getFileContents("/usr/local/ispconfig/server/lib/config.inc.php");
                $re = '/\$conf\[\'server_id\'\]\W=\W\'[0-9]+\';/m';

                $config_content = preg_replace($re, "\$conf['server_id'] = '1';", $config_content);
                $ssh->putFileContents("/usr/local/ispconfig/server/lib/config.inc.php", $config_content);
            }
        }
    }

    public function printSuccessMessage() {
        global $ssh;

        $this->swriteln();
        $this->swriteln();
        $this->swriteln("=============================================================");
        $this->swriteln();
        $this->swriteln("Old (source) server:");
        $this->swriteln($this->local_server_hostname);
        $this->swriteln();
        $this->swriteln("New (target) server:");
        $this->swriteln($this->target_server_hostname);
        $this->swriteln();
        $this->swriteln("Old (source) IP:");
        $this->swriteln($this->local_server_ip);
        $this->swriteln();
        $this->swriteln("New (target) IP:");
        $this->swriteln($this->target_server_ip);
        $this->swriteln();
        $this->swriteln("=============================================================");
        $log_file = $this->getScreenLog();

        $this->swriteln("You can find the screen-log here: {$log_file}");


        $this->fixMonitoringMessages();
        $this->swriteln();
        $this->generateHostList();

        exec("rm -rf " . $this->mysql_defaults_extra_file);
        $ssh->exec("rm -rf " . $this->mysql_defaults_extra_file);

        $this->swriteln();
    }

    private function fixMonitoringMessages() {
        global $ssh;

        if($this->getInputLogEntry("mode") == "single" || $this->getInputLogEntry("mode") == "second") {
            $this->swriteln("Preserve monitoring errors...");
            $files = [
                "/usr/local/lib/nagios/plugins/check_ispconfig_cron_deadjob",
                "/usr/local/lib/nagios/plugins/check_ispconfig_server_sh",
                "/usr/local/lib/nagios/plugins/check_ispconfig_cron_sh",
                "/usr/lib/nagios/plugins/check_mailq"
            ];
            $new_content = <<<TXT
#!/bin/bash -

echo "OK - disabled by ISPCopy"
exit 0
TXT;
            foreach($files as $file) {
                if(file_exists($file)) {
                    exec("mv -f $file $file.bak");
                    file_put_contents($file, $new_content);
                    chmod($file, 0755);
                }
            }
        }
    }

    private function getScreenLog() {
        $screen_log_file = "{$this->getLogDir()}/ispcopy_screen_";
        $log_file_date = date("Y-m-d H:i:s");
        while (!file_exists($screen_log_file . date("Ymd-Hi", strtotime($log_file_date)) . ".log")){
            $log_file_date = date("Y-m-d H:i:s", strtotime($log_file_date . " -1 minute"));
        }
        $screen_log_file .= date("Ymd-Hi", strtotime($log_file_date)) . ".log";
        return $screen_log_file;
    }

    /**
     * @param $server (source|target)
     */
    public function runServerSh($server = "source") {
        global $ssh;
        if($server == "source") {
            $this->swriteln("############ Run server.sh on source server ############");
            exec("server.sh", $output);
            if(strpos($output[0], "There is already an instance of server.php running") != false) {
                $this->swriteln(implode("\n", $output));
                sleep(20);
                $this->runServerSh($server);
            }
            $this->swriteln(implode("\n", $output));
        } elseif ($server == "target") {
            $this->swriteln("############ Run server.sh on target server ############");
            $result = $ssh->exec("server.sh");
            if(strpos($result, "There is already an instance of server.php running") != false) {
                $this->swriteln($result);
                sleep(20);
                $this->runServerSh($server);
            }
            $this->swriteln($result);
        }
    }

    public function disableServerShOnTarget() {
        global $ssh;

        $ssh->exec("crontab -l > /root/sys_crons.tmp");
        $crons = explode("\n", $ssh->getFileContents("/root/sys_crons.tmp"));
        for ($i = 0; $i < count($crons); $i++) {
            if (substr(trim($crons[$i]), 0, 1) != "#" && strpos($crons[$i], "server.sh") != false) {
                $crons[$i] = "#" . $crons[$i];
            }
        }
        $ssh->putFileContents("/root/sys_crons.tmp", implode("\n", $crons));
        $ssh->exec("crontab /root/sys_crons.tmp");
    }

    public function enableServerShOnTarget() {
        global $ssh;

        $ssh->exec("crontab -l > /root/sys_crons.tmp");
        $crons = explode("\n", $ssh->getFileContents("/root/sys_crons.tmp"));
        for ($i = 0; $i < count($crons); $i++) {
            if (substr(trim($crons[$i]), 0, 1) == "#" && strpos($crons[$i], "server.sh") != false) {
                $crons[$i] = substr($crons[$i], 1);
            }
        }
        $ssh->putFileContents("/root/sys_crons.tmp", implode("\n", $crons));
        $ssh->exec("crontab /root/sys_crons.tmp");
    }

    /**
     * Prints a message and waits for enter befor send a die() to end isp_copy.
     *
     * @param $msg
     * @return void
     */
    public function end_copy($msg = "An error occured, the application will be closed after pressing enter.") {
        $this->free_query($msg, "");
        die();
    }
}