<?php

define('ISPC_RSYNC_PROGRESS', 1);

class ssh_lib extends lib {
	
	private $hostname;
	private $port = 22;
	private $username = 'root';
	private $use_identity_file = true;
	private $identity_file = "/root/.ssh/id_rsa_migration";
	private $local_rsync_version = false;
	private $remote_rsync_version = false;
	
	public function __construct($hostname = '', $port = 22, $username = 'root', $identity_file = null) {
		$this->hostname = $hostname;
		$this->port = $port;
		$this->username = $username;
		if($identity_file && file_exists($identity_file)) {
			$this->identity_file = $identity_file;
		}
        if(!file_exists($this->identity_file)) {
            echo("Create migration SSH key\n");
            exec("ssh-keygen -q -N '' -b 4096 -t rsa -f {$this->identity_file}");
        }
        chmod($this->identity_file, 0600);
        chmod("{$this->identity_file}.pub", 0600);
	}
	
	
	protected function log($message, $priority = 0) {
		// Priority: 0 = debug, 1 = warn, 2 = error.
		if(defined('APP_NAME') && APP_NAME == 'ISPCopy') {
			global $icp;
			$icp->log($message, $priority);
		} elseif(class_exists('ISPConfigLog')) {
			ISPConfigLog::info($message);
		} else {
			// TODO: Implement logging
			if($priority >= $this->getLogLevel()) {
				$this->do_echo($message."\n");
			}
		}
	}
	
	protected function do_echo($msg) {
		echo $msg;
	}
	
	public function setHostname($hostname) {
		$this->hostname = $hostname;
	}
	
	public function setPort($port) {
		$this->port = $port;
	}
	
	public function setUsername($username) {
		$this->username = $username;
	}

	public function getHostname() {
		return $this->hostname;
	}
	
	public function getPort() {
		return $this->port;
	}
	
	public function getUsername() {
		return $this->username;
	}

    public function getIdentityFile() {
        return $this->identity_file;
    }

	private function checkSettings() {
		if($this->hostname == '') {
			$this->log('Hostname not set',2);
			return false;
		}
		return true;
	}


	//* Test the SSH connection
	public function testConnect() {
		
		// Test the settings first
		if(!$this->checkSettings()) return false;
		
		// Check connection without identity file
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.escapeshellarg($this->username.'@'.$this->hostname).' cat /etc/hosts > /dev/null 2>&1',$output,$return_var);
		if($return_var === 0) {
			$this->use_identity_file = true;
			return true;
		} else {
			// Do an SSH test connect with identity file
			exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes -i '.$this->identity_file.' '.escapeshellarg($this->username.'@'.$this->hostname).' cat /etc/hosts > /dev/null 2>&1',$output,$return_var);
			if($return_var === 0) {
				$this->use_identity_file = true;
				return true;
			} else {
				return false;
			}
		}
	}
	
	//* Install a cert based SSH login
	public function installSSHCertDEP() {
			if(!file_exists($this->identity_file)) {
				$this->do_echo("Creating SSH key.\n");
				passthru('ssh-keygen -q -N \'\' -b 4096 -t rsa -f '.$this->identity_file);
			}
			$this->do_echo("Now we transfer the key to the target server.\n");
			$check = exec('ssh-copy-id 2>&1 | head -n 1');
			if(strpos($check, '-p port') !== false) {
				passthru('ssh-copy-id -i '.$this->identity_file.' -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 '. escapeshellarg($this->username.'@'.$this->hostname));
			} else {
				passthru('ssh-copy-id -i '.$this->identity_file.' ' . escapeshellarg('-p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 '.$this->username.'@'.$this->hostname));
			}
			$this->use_identity_file = true;
	}

    public function installSSHCert($username = "") {
        $this->do_echo("Copy root ssh key.");
        $tmp = $this->username;
        if(!empty($username)) {
            $this->username = $username;
        }
        exec("ssh-keygen -f \"/root/.ssh/known_hosts\" -R \"{$this->hostname}\" >> /dev/null 2>&1");
        $pub_key = trim(file_get_contents("{$this->identity_file}.pub"));
        $this->exec("echo '{$pub_key}' | sudo tee -a /root/.ssh/authorized_keys2");
        exec("ssh-keygen -f \"/root/.ssh/known_hosts\" -R \"{$this->hostname}\" >> /dev/null 2>&1");
        $this->username = $tmp;
    }

    public function create_deafaults_extra_file() {
        if(!$this->isFile($this->mysql_defaults_extra_file)) {
            $tmpfile = tempnam(sys_get_temp_dir(), "mysql-");
            $this->getFile('/usr/local/ispconfig/server/lib/mysql_clientdb.conf', $tmpfile);
            include $tmpfile;
            $content = <<<TXT
[client]
host=$clientdb_host
user=$clientdb_user
password=$clientdb_password
TXT;
            if(!$this->isDir($this->getLogDir())) {
                $this->mkdir($this->getLogDir());
            }
            $this->putFileContents($this->mysql_defaults_extra_file, $content);
            unlink($tmpfile);
        }
    }

	//* get a file from the remote server
	public function getFile($remote, $local) {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "cat '.escapeshellarg($remote).'" > '.escapeshellarg($local),$output,$return_var);
		if($return_var === 0) {
			// check if the file was really transferred
			if(file_exists($local)) {
				$this->log('File '.$local.' successfully transferred.',0);
				return true;
			} else {
				$this->log('Error: File '.$local.' transfer failed.',1);
				return false;
			}
		} else {
			$this->log('SSH Error: File '.$local.' could not be transferred.',1);
			return false;
		}
	}
	
	//* Send a file to the remote server
	public function putFile($local, $remote, $compress = false, $is_dir = false) {
		if(!file_exists($local)) {
			$this->log('Error: File '.$local.' does not exist.',1);
			return false;
		}
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('scp -p' . ($compress ? ' -C' : '') . ($is_dir ? ' -r' : '') . ' -P ' . $this->port. ' -o StrictHostKeyChecking=no -o ConnectTimeout=5 -o BatchMode=yes '.$identity.' '.escapeshellarg($local) . ' ' . escapeshellarg($this->username.'@'.$this->hostname . ':' . $remote), $output, $return_var);
		//exec('cat '.escapeshellarg($local).' | ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "cat > '.escapeshellarg($remote).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('File '.$remote.' successfully transferred.',0);
			return true;
		} else {
			$this->log('Error: File '.$remote.' transfer failed.',1);
			return false;
		}
	}
	
	//* Get a file content from remote server
	public function getFileContents($remote) {
		$local = tempnam(sys_get_temp_dir(), "ssh");
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "cat '.escapeshellarg($remote).'" > '.escapeshellarg($local),$output,$return_var);
		if($return_var === 0) {
			// check if the file was really transferred
			if(file_exists($local)) {
				$this->log('File '.$local.' successfully transferred.',0);
				$content = file_get_contents($local);
				unlink($local);
				return $content;
			} else {
				$this->log('Error: File '.$local.' transfer failed.',1);
				return false;
			}
		} else {
			$this->log('SSH Error: File '.$local.' could not be transferred.',1);
			return false;
		}
	}
	
	//* Write file on remote server
	public function putFileContents($remote,$content,$mode = 'replace') {
		$local = tempnam(sys_get_temp_dir(), "ssh");
		file_put_contents($local,$content);
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		if($mode == 'replace') {
			// Replace the file
			$copycmd = '>';
		} else {
			// append content to file
			$copycmd = '>>';
		}
		exec('cat '.escapeshellarg($local).' | ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "cat '.$copycmd.' '.escapeshellarg($remote).'"',$output,$return_var);
		unlink($local);
		if($return_var === 0) {
			$this->log('File '.$remote.' successfully transferred.',0);
			return true;
		} else {
			$this->log('Error: File '.$remote.' transfer failed.',1);
			return false;
		}
	}
	
	//* Copy file on remote server
	public function cp($from_path, $to_path, $options = '-prf') {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "cp '.$options.' '.escapeshellarg($from_path).' '.escapeshellarg($to_path).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('File '.$from_path.' successfully copied.',0);
			return true;
		} else {
			$this->log('Error: File '.$from_path.' copy failed.',1);
			return false;
		}
	}
	
	//* move file on remote server
	public function mv($from_path, $to_path, $options = '-f') {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "mv '.$options.' '.escapeshellarg($from_path).' '.escapeshellarg($to_path).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('File '.$from_path.' successfully moved.',0);
			return true;
		} else {
			$this->log('Error: File '.$from_path.' move failed.',1);
			return false;
		}
	}
	
	//* Check if a file exists
	public function isFile($path) {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "test -e '.escapeshellarg($path).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('File '.$path.' exists.',0);
			return true;
		} else {
			$this->log('File '.$path.' does not exist.',0);
			return false;
		}
	}
	
	
	public function mkdir($path, $mode = '0750') {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "mkdir -p -m ' . escapeshellarg($mode) . ' '.escapeshellarg($path).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('Directory '.$path.' created.',0);
			return true;
		} else {
			$this->log('Directory '.$path.' not created.',0);
			return false;
		}
	}
	
	//* Check if a directory exists
	public function isDir($path) {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' "test -d '.escapeshellarg($path).'"',$output,$return_var);
		if($return_var === 0) {
			$this->log('Directory '.$path.' exists.',0);
			return true;
		} else {
			$this->log('Directory '.$path.' does not exist.',0);
			return false;
		}
	}
	
	//* Execute command on the remote server
	public function exec($cmd, $returncodes_ok = array(), &$return_code = null) {
		$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
		exec('ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=5 -oBatchMode=yes '.$identity.' '.escapeshellarg($this->username.'@'.$this->hostname).' '.escapeshellarg($cmd),$output,$return_var);
		if($return_var === 0 || (!empty($returncodes_ok) && is_array($returncodes_ok) && in_array($return_var, $returncodes_ok, true))) {
			$this->log('Successfully executed command '.$cmd,0);
			$return_code = $return_var;
			return implode("\n",$output);
		} else {
			$this->log('Command '.$cmd.' failed with code ' . $return_var . '.',0);
			$return_code = $return_var;
			return false;
		}
	}
	
	//* Test MySQL login credentials on remote server
	public function testDBLogin($hostname,$username,$password) {
		if($password != '') {
			$result = $this->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -e "select now()" 2> /dev/null');
		} else {
			$result = $this->exec('mysql --defaults-extra-file='.$this->mysql_defaults_extra_file.' -e "select now()" 2> /dev/null');
		}
		if($result === false) {
			return false;
		} else {
			return true;
		}
	}
	
	
	//* Copy files by rsync or scp
	public function copyFiles($from_path, $to_path, $mode = 'rsync', $delete = true, $flags = false) {
		// rsync mode
		if($mode == 'rsync') {
			if(!$this->local_rsync_version) {
				exec('rsync --version 2>/dev/null', $output, $return_var);
				if($return_var === 0 && preg_match('/rsync\s+version\s+(\d+(?:\.\d+)*)\s/i', $output[0], $match)) {
					$this->local_rsync_version = $match[1];
				}
			}
			if(!$this->remote_rsync_version) {
				$tmp = $this->exec('rsync --version 2>/dev/null');
				if($tmp !== false && preg_match('/rsync\s+version\s+(\d+(?:\.\d+)*)\s/i', $tmp, $match)) {
					$this->remote_rsync_version = $match[1];
				}
			}
			
			if($flags & ISPC_RSYNC_PROGRESS) {
				// we need to check capability
				if(version_compare($this->local_rsync_version, '3.1.0', '<') || version_compare($this->remote_rsync_version, '3.1.0', '<')) {
					$flags = $flags & ~ISPC_RSYNC_PROGRESS;
					$this->log('Cannot display rsync progress as either local or remote version is lower than 3.1.0.', 0);
				}
			}
			$delete_option = ($delete == true)?'--delete-during':'';
			$identity = ($this->use_identity_file)?'-i '.$this->identity_file:'';
			$add_options = '';
			if($flags & ISPC_RSYNC_PROGRESS) $add_options .= ' --info=progress2';
			$cmd = 'rsync -a';
			if(!($flags & ISPC_RSYNC_PROGRESS)) $cmd .= 'q';
			
			// do not copy Jailkits, see https://git.timmehosting.de/devops/ispconfig/ispconfig_migration-tool/-/issues/13
			//$excludes = " --exclude 'clients/client*/web*/.bash_history' --exclude 'clients/client*/web*/.bashrc' --exclude 'clients/client*/web*/bin' --exclude 'clients/client*/web*/dev' --exclude 'clients/client*/web*/etc' --exclude 'clients/client*/web*/lib' --exclude 'clients/client*/web*/lib64' --exclude 'clients/client*/web*/opt' --exclude 'clients/client*/web*/.profile' --exclude 'clients/client*/web*/usr' --exclude 'clients/client*/web*/var'"; // alternative Schreibweise, funktioniert
			//$excludes = ' --exclude="clients/client*/web*/.bash_history" --exclude="clients/client*/web*/.bashrc" --exclude="clients/client*/web*/bin" --exclude="clients/client*/web*/dev" --exclude="clients/client*/web*/etc" --exclude="clients/client*/web*/lib" --exclude="clients/client*/web*/lib64" --exclude="clients/client*/web*/opt" --exclude="clients/client*/web*/.profile" --exclude="clients/client*/web*/usr" --exclude="clients/client*/web*/var"'; // alternative Schreibweise, funktioniert
			$excludes = " --exclude='clients/client*/web*/.bash_history' --exclude='clients/client*/web*/.bashrc' --exclude='clients/client*/web*/bin' --exclude='clients/client*/web*/dev' --exclude='clients/client*/web*/etc' --exclude='clients/client*/web*/lib' --exclude='clients/client*/web*/lib64' --exclude='clients/client*/web*/opt' --exclude='clients/client*/web*/.profile' --exclude='clients/client*/web*/usr' --exclude='clients/client*/web*/var' --exclude='apps/phpmyadmin' --exclude='apps/webmail'";
			
			$cmd .= ' '.$delete_option.$add_options.$excludes.' -e "ssh -p '.$this->port.' -oStrictHostKeyChecking=no -oConnectTimeout=10 -oBatchMode=yes '.$identity.'" '.escapeshellarg($from_path).' '.escapeshellarg($this->username.'@'.$this->hostname.':'.$to_path);
			if($flags & ISPC_RSYNC_PROGRESS) $cmd .= ' 2>/dev/null';
			else $cmd .= ' >/dev/null 2>&1';
			if($flags & ISPC_RSYNC_PROGRESS) {
				print "\n";
				passthru($cmd, $return_var);
			} else {
				exec($cmd,$output,$return_var);
			}
			if($return_var === 0 || $return_var === 23 || $return_var === 24) {
				$this->log('Rsync copy from '.$from_path.' to ' . $to_path . ' successful.',0);
				return true;
			} else {
				$this->log('Error: Rsync copy from '.$from_path.' to ' . $to_path . ' failed. Return value '.$return_var.'.',1);
				$this->log('Command was: ' . $cmd, 0);
				return false;
			}
		}
		// TODO: scp mode
		return false;
	}
	
	
}

/*

SSH Error codes

	0
 	Success 
   	1 
 	Undetermined error in file copy 
   	3 
 	Destination is not directory, but it should be 
   	4 
 	Connecting to host failed 
   	5 
 	Connection lost for some reason 
   	6 
 	File does not exist 
   	7 
 	No permission to access file 
   	8 
 	Undetermined error from sshfilexfer 
   	9 
 	File transfer protocol mismatch 
   	255 
 	Error occurred in SSH 


RSYNC Exit values

EXIT VALUES 
 ������0 �����Success 

 ������1 �����Syntax or usage error 

 ������2 �����Protocol incompatibility 

 ������3 �����Errors selecting input/output files, dirs 

 ������4 �����Requested �action not supported: an attempt was made to manipulate 64-bit files on a platform that can- 
 �������������not support them; or an option was specified that is supported by the client and not by the server. 

 ������5 �����Error starting client-server protocol 

 ������6 �����Daemon unable to append to log-file 

 ������10 ����Error in socket I/O 

 ������11 ����Error in file I/O 

 ������12 ����Error in rsync protocol data stream 

 ������13 ����Errors with program diagnostics 

 ������14 ����Error in IPC code 

 ������20 ����Received SIGUSR1 or SIGINT 

 ������21 ����Some error returned by waitpid() 

 ������22 ����Error allocating core memory buffers 

 ������23 ����Partial transfer due to error 

 ������24 ����Partial transfer due to vanished source files 

 ������25 ����The --max-delete limit stopped deletions 

 ������30 ����Timeout in data send/receive

 ������35 ����Timeout waiting for daemon connection





*/